###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os
import subprocess
import sys

from cmflib.cli.command import CmdBase
from cmflib.dvc_wrapper import (
    git_quiet_init,
    git_checkout_new_branch,
    git_initial_commit,
    git_add_remote,
    check_git_repo,
    dvc_quiet_init,
    dvc_add_remote_repo,
    dvc_add_attribute,
)
from cmflib.utils.cmf_config import CmfConfig
from cmflib.utils.helper_functions import is_git_repo


class CmdInitMinioS3(CmdBase):
    def run(self):
        # Reading CONFIG_FILE variable
        cmf_config = os.environ.get("CONFIG_FILE", ".cmfconfig")
        # checking if config file exists
        if not os.path.exists(cmf_config):
            # writing default value to config file
            attr_dict = {}
            attr_dict["server-ip"] = "http://127.0.0.1:80"
            CmfConfig.write_config(cmf_config, "cmf", attr_dict)

        # if user gave --cmf-server-url, override the config file
        if self.args.cmf_server_url:
            attr_dict = {}
            attr_dict["server-ip"] = self.args.cmf_server_url
            CmfConfig.write_config(cmf_config, "cmf", attr_dict, True)

        # read --neo4j details and add to the exsting file
        if self.args.neo4j_user and self.args.neo4j_password and self.args.neo4j_uri:
            attr_dict = {}
            attr_dict["user"] = self.args.neo4j_user
            attr_dict["password"] = self.args.neo4j_password
            attr_dict["uri"] = self.args.neo4j_uri
            CmfConfig.write_config(cmf_config, "neo4j", attr_dict, True)
        elif (
            not self.args.neo4j_user
            and not self.args.neo4j_password
            and not self.args.neo4j_uri
        ):
            pass
        else:
            return "ERROR: Provide user, password and uri for neo4j initialization."
        output = is_git_repo()
        if not output:
            branch_name = "master"
            print("Starting git init.")
            git_quiet_init()
            git_checkout_new_branch(branch_name)
            git_initial_commit()
            git_add_remote(self.args.git_remote_url)
            print("git init complete.")

        print("Starting cmf init.")
        dvc_quiet_init()
        repo_type = "minio"
        output = dvc_add_remote_repo(repo_type, self.args.url)
        if not output:
            return "cmf init failed."
        print(output)
        dvc_add_attribute(repo_type, "endpointurl", self.args.endpoint_url)
        dvc_add_attribute(repo_type, "access_key_id", self.args.access_key_id)
        dvc_add_attribute(repo_type, "secret_access_key", self.args.secret_key)
        return "cmf init complete."


def add_parser(subparsers, parent_parser):
    HELP = "Initialises Minio S3 bucket as artifact repository."

    parser = subparsers.add_parser(
        "minioS3",
        parents=[parent_parser],
        description="This command initialises Minio S3 bucket as artifact repository for CMF.",
        help=HELP,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "--url",
        required=True,
        help="Specify Minio S3 bucket url.",
        metavar="<url>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--endpoint-url",
        required=True,
        help="Specify endpoint url which is used to access Minio's locally/remotely running UI.",
        metavar="<endpoint_url>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--access-key-id",
        required=True,
        help="Specify Access Key Id.",
        metavar="<access_key_id>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--secret-key",
        required=True,
        help="Specify Secret Key.",
        metavar="<secret_key>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--git-remote-url",
        required=True,
        help="Specify git repo url. eg: https://github.com/XXX/example.git",
        metavar="<git_remote_url>",
        default=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--cmf-server-url",
        help="Specify cmf-server URL",
        metavar="<cmf_server_url>",
        default="http://127.0.0.1:80",
    )

    parser.add_argument(
        "--neo4j-user",
        help="Specify neo4j user.",
        metavar="<neo4j_user>",
        # default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--neo4j-password",
        help="Specify neo4j password.",
        metavar="<neo4j_password>",
        # default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--neo4j-uri",
        help="Specify neo4j uri.eg bolt://localhost:7687",
        metavar="<neo4j_uri>",
        # default=argparse.SUPPRESS,
    )

    parser.set_defaults(func=CmdInitMinioS3)



import itertools
import re
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout
import pandas as pd
from cmflib import cmfquery
import dvc
import json
import random
import warnings

warnings.filterwarnings("ignore")

def truncate_artifact_name(my_str):
    if "/" in my_str:
        my_str = my_str.split("/")[-1]
    temp = my_str.split(":")
    temp[1] = temp[1][:4]
    temp=":".join(temp)
    return temp

def query_artifact_lineage_d3force(mlmd_path, pipeline_name, dict_of_art_ids):
    art_name_id = {}
    artifact_name_list = []
    query = cmfquery.CmfQuery(mlmd_path)
    node_id_name_list = []
    pipeline_id=query.get_pipeline_id(pipeline_name)
    for type_, df in dict_of_art_ids[pipeline_name].items():
        for index, row in df.iterrows():
            artifact_name_list.append(row["name"])
            art_name_id[row["name"]] = row["id"]
            result="#48D1CC" if type_ == "Dataset" else "#2E8B57" if type_ =="Model" else "#FF8C00"
            node_id_name_list.append({"name": truncate_artifact_name(row["name"]), "id": row["id"], "color": result})
    link_list = []
    for art_name in artifact_name_list:
        immediate_child_artifacts = query.get_one_hop_child_artifacts(art_name, pipeline_id)
        if immediate_child_artifacts.empty == True:
            pass
        else:
            for i in list(set(immediate_child_artifacts["name"])):
                link = {}
                srce_art_id = art_name_id[art_name]
                trgt_art_id = art_name_id[i]
                link["source"] = int(srce_art_id)
                link["target"] = (trgt_art_id)
                link_list.append(link)

    new_list = pd.DataFrame(
    link_list
    ).drop_duplicates().to_dict('records')
    data = {
        "nodes" : node_id_name_list,
        "links" : new_list
    }
    return data

#print(query_visualization("/home/chobey/repair_lineage/testenv/example-get-started/mlmd","Test-env"))



from cmflib import cmfquery, cmf_merger
import pandas as pd
import json
import os
from server.app.query_artifact_lineage_d3force import query_artifact_lineage_d3force
from server.app.query_list_of_executions import query_list_of_executions
from fastapi.responses import FileResponse


async def get_model_data(mlmdfilepath, modelId):
    '''
      This function retrieves the necessary model data required for generating a model card.

      Arguments:
        mlmdfilepath (str): The file path to the metadata.
        modelId (int): The ID of the model for which data is required.

      Returns:
        This function returns a tuple of DataFrames containing the following:

        model_data_df (DataFrame): Metadata related to the model itself.
        model_exe_df (DataFrame): Metadata of the executions in which the specified modelId was an input or output.
        model_input_df (DataFrame): Metadata of input artifacts that led to the creation of the model.
        model_output_df (DataFrame): Metadata of artifacts that used the model as an input.
        The returned DataFrames provide comprehensive metadata for the specified model, aiding in the creation of detailed and accurate model cards.
    '''
    query = cmfquery.CmfQuery(mlmdfilepath)
    pd.set_option('display.max_columns', None)
    model_data_df = pd.DataFrame()
    model_exe_df = pd.DataFrame()
    model_input_df = pd.DataFrame()
    model_output_df = pd.DataFrame()

    # get name from id
    modelName = ""
    model_data_df = query.get_all_artifacts_by_ids_list([modelId])
    # if above dataframe is not empty, we have the dataframe for given modelId with full model related details
    if model_data_df.empty:
        return model_data_df, model_exe_df, model_input_df, model_output_df
    # However following check is done, in case, variable 'modelId' is not an ID for model artifact
    modelType = model_data_df['type'].tolist()[0]
    if not modelType == "Model":
        # making model_data_df empty
        model_data_df = pd.DataFrame()
        return model_data_df, model_exe_df, model_input_df, model_output_df


    # extracting modelName
    modelName = model_data_df['name'].tolist()[0]

    # model's executions data with props and custom props
    exe_df = query.get_all_executions_for_artifact(modelName)
    exe_ids = []
    if not exe_df.empty:
        exe_df.drop(columns=['execution_type_name', 'execution_name'], inplace=True)
        exe_ids = exe_df['execution_id'].tolist()


    if not exe_ids:
         return model_data_df, model_exe_df, model_input_df, model_output_df
    model_exe_df = query.get_all_executions_by_ids_list(exe_ids)
    model_exe_df.drop(columns=['Python_Env', 'Git_Start_Commit', 'Git_End_Commit'], inplace=True)

    in_art_ids =  []
    # input artifacts
    # it is usually not a good practice to use functions starting with _ outside of the file they are defined .. should i change??
    in_art_ids.extend(query._get_input_artifacts(exe_ids))
    if modelId in in_art_ids:
        in_art_ids.remove(modelId)
    model_input_df = query.get_all_artifacts_by_ids_list(in_art_ids)

    out_art_ids = []
    # output artifacts
    out_art_ids.extend(query._get_output_artifacts(exe_ids))
    if modelId in out_art_ids:
        out_art_ids.remove(modelId)
    model_output_df = query.get_all_artifacts_by_ids_list(out_art_ids)

    return model_data_df, model_exe_df, model_input_df, model_output_df

async def get_executions(mlmdfilepath, pipeline_name, exe_ids):
    '''
    Args:
     mlmdfilepath: mlmd file path.
     pipeline_name: name of the pipeline.
     exe_ids: list of execution ids.

    Returns:
     returns dataframe of executions using execution_ids.
    '''
    query = cmfquery.CmfQuery(mlmdfilepath)
    df = pd.DataFrame()
    executions = query.get_all_executions_by_ids_list(exe_ids)
    df = pd.concat([df, executions], sort=True, ignore_index=True)
    #df=df.drop('name',axis=1)
    return df


async def get_all_exe_ids(mlmdfilepath, pipeline_name: str = None):
    '''
    Returns:
        returns a dictionary which has pipeline_name as key and dataframe which includes {id,Execution_uuid,Context_Type,Context_id} as value.
    '''
    query = cmfquery.CmfQuery(mlmdfilepath)
    execution_ids = {}
    executions = pd.DataFrame()    # df is emptied to store execution ids for next pipeline.
    if pipeline_name:
        executions = query.get_all_executions_in_pipeline(pipeline_name)
        if not executions.empty:
            execution_ids[pipeline_name] = executions[['id', 'Context_Type', 'Execution_uuid', 'Context_ID']]
        else:
            execution_ids[pipeline_name] = pd.DataFrame()
    else:
        names = query.get_pipeline_names()
        for name in names:
            executions = pd.DataFrame()    # df is emptied to store execution ids for next pipeline.
            executions = query.get_all_executions_in_pipeline(name)
            # check if df is empty return just pipeline_name: {}
            # if df is not empty return dictionary with pipeline_name as key
            # and df with id, context_type, uuid, context_ID as value.
            if not executions.empty:
                execution_ids[name] = executions[['id', 'Context_Type', 'Execution_uuid', 'Context_ID']]
            else:
                execution_ids[name] = pd.DataFrame()
    return execution_ids


async def get_all_artifact_ids(mlmdfilepath, execution_ids, pipeline_name: str = None):
    # following is a dictionary of dictionaries
    # First level dictionary key is pipeline_name
    # First level dicitonary value is nested dictionary
    # Nested dictionary key is type i.e. Dataset, Model, etc.
    # Nested dictionary value is a pandas df with id and artifact name
    artifact_ids = {}
    query = cmfquery.CmfQuery(mlmdfilepath)
    artifacts = pd.DataFrame()
    if pipeline_name:
        if not execution_ids.get(pipeline_name).empty:
            exe_ids = execution_ids[pipeline_name]['id'].tolist()
            artifacts = query.get_all_artifacts_for_executions(exe_ids)
            #acknowledging pipeline exist even if df is empty. 
            if artifacts.empty:
                artifact_ids[pipeline_name] = pd.DataFrame()   # { pipeline_name: {empty df} }
            else:
                artifact_ids[pipeline_name] = {}
                for art_type in artifacts['type']:
                    filtered_values = artifacts.loc[artifacts['type'] == art_type, ['id', 'name']]
                    artifact_ids[pipeline_name][art_type] = filtered_values
        # if execution_ids is empty then create dictionary with key as pipeline name
        # and value as empty df
        else:
            artifact_ids[pipeline_name] = pd.DataFrame()
    else:
        names = query.get_pipeline_names()
        for name in names:
            if not execution_ids.get(name).empty:
                exe_ids = execution_ids[name]['id'].tolist()
                artifacts = query.get_all_artifacts_for_executions(exe_ids)
                #acknowledging pipeline exist even if df is empty. 
                if artifacts.empty:
                    artifact_ids[name] = pd.DataFrame()   # { pipeline_name: {empty df} }
                else:
                    artifact_ids[name] = {}
                    for art_type in artifacts['type']:
                        filtered_values = artifacts.loc[artifacts['type'] == art_type, ['id', 'name']]
                        artifact_ids[name][art_type] = filtered_values
            # if execution_ids is empty then create dictionary with key as pipeline name
            # and value as empty df
            else:
                artifact_ids[name] = pd.DataFrame()
    return artifact_ids


async def get_artifacts(mlmdfilepath, pipeline_name, art_type, artifact_ids):
    query = cmfquery.CmfQuery(mlmdfilepath)
    df = pd.DataFrame()
    if (query.get_pipeline_id(pipeline_name) != -1):
        df = query.get_all_artifacts_by_ids_list(artifact_ids)
        if len(df) == 0:
            return
        df = df.drop_duplicates()
        art_names = df['name'].tolist()
        name_dict = {}
        name_list = []
        exec_type_name_list = []
        exe_type_name = pd.DataFrame()
        for name in art_names:
            executions = query.get_all_executions_for_artifact(name)
            exe_type_name = pd.concat([exe_type_name, executions], ignore_index=True)
            execution_type_name = exe_type_name["execution_type_name"].drop_duplicates().tolist()
            execution_type_name = [str(element).split('"')[1] for element in execution_type_name]
            execution_type_name_str = ',\n '.join(map(str, execution_type_name))
            name_list.append(name)
            exec_type_name_list.append(execution_type_name_str)
        name_dict['name'] = name_list
        name_dict['execution_type_name'] = exec_type_name_list
        name_df = pd.DataFrame(name_dict)
        merged_df = df.merge(name_df, on='name', how='left')
        merged_df['name'] = merged_df['name'].apply(lambda x: x.split(':')[0] if ':' in x else x)
        merged_df = merged_df.loc[merged_df["type"] == art_type]
        result = merged_df.to_json(orient="records")
        tempout = json.loads(result)
        return tempout

def get_artifact_types(mlmdfilepath):
    query = cmfquery.CmfQuery(mlmdfilepath)
    artifact_types = query.get_all_artifact_types()
    return artifact_types

async def create_unique_executions(server_store_path, req_info):
    mlmd_data = json.loads(req_info["json_payload"])
    pipelines = mlmd_data["Pipeline"]
    pipeline = pipelines[0]
    pipeline_name = pipeline["name"]
    executions_server = []
    list_executions_exists = []
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        executions = query.get_all_executions_in_pipeline(pipeline_name)
        for i in executions.index:
            for uuid in executions['Execution_uuid'][i].split(","):
                executions_server.append(uuid)
        executions_client = []
        for i in mlmd_data['Pipeline'][0]["stages"]:  # checks if given execution_id present in mlmd
            for j in i["executions"]:
                if j['name'] != "": #If executions have name , they are reusable executions
                    continue       #which needs to be merged in irrespective of whether already
                                   #present or not so that new artifacts associated with it gets in.
                if 'Execution_uuid' in j['properties']:
                    for uuid in j['properties']['Execution_uuid'].split(","):
                        executions_client.append(uuid)
                else:
                    # mlmd push is failed here
                    status="version_update"
                    return status
        if executions_server != []:
            list_executions_exists = list(set(executions_client).intersection(set(executions_server)))
        for i in mlmd_data["Pipeline"]:
            for stage in i['stages']:
                for cmf_exec in stage['executions'][:]:
                    uuids = cmf_exec["properties"]["Execution_uuid"].split(",")
                    for uuid in uuids:
                        if uuid in list_executions_exists:
                            stage['executions'].remove(cmf_exec)

        for i in mlmd_data["Pipeline"]:
            i['stages']=[stage for stage in i['stages'] if stage['executions']!=[]]
    for i in mlmd_data["Pipeline"]:
        if len(i['stages']) == 0 :
            status="exists"
        else:
            cmf_merger.parse_json_to_mlmd(
                json.dumps(mlmd_data), "/cmf-server/data/mlmd", "push", req_info["id"]
            )
            status='success'
    return status


async def get_mlmd_from_server(server_store_path: str, pipeline_name: str, exec_id: str):
    query = cmfquery.CmfQuery(server_store_path)
    execution_flag = 0
    json_payload = None
    df = pd.DataFrame()
    if(query.get_pipeline_id(pipeline_name)!=-1):  # checks if pipeline name is available in mlmd
        if exec_id != None:
            exec_id = int(exec_id)
            df = query.get_all_executions_by_ids_list([exec_id])
            if df.empty:
                json_payload = "no_exec_id"
                return json_payload
        json_payload = query.dumptojson(pipeline_name, exec_id)
    return json_payload


async def get_lineage_data(server_store_path,pipeline_name,type,dict_of_art_ids,dict_of_exe_ids):
    query = cmfquery.CmfQuery(server_store_path)
    if type=="Artifacts":
        lineage_data = query_artifact_lineage_d3force(server_store_path, pipeline_name, dict_of_art_ids)
        '''
        returns dictionary of nodes and links for artifact lineage.
        lineage_data= {
                       nodes:[],
                       links:[]
                      }
        '''
    elif type=="Execution":
        lineage_data = query_list_of_executions(server_store_path, pipeline_name, dict_of_art_ids, dict_of_exe_ids)
        '''
        returns list of execution types for specific pipeline.
        '''
    else:
        lineage_data = query_visualization_ArtifactExecution(server_store_path, pipeline_name)
    return lineage_data




###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
from neo4j import GraphDatabase
import typing as t
import re
from ml_metadata.proto import metadata_store_pb2 as mlpb


class GraphDriver:

    def __init__(self, uri, user, password):
        self.driver = GraphDatabase.driver(uri, auth=(user, password))
        self.pipeline_id = None
        self.stage_id = None
        self.execution_id = None

    def close(self):
        self.driver.close()

    def create_pipeline_node(self, name: str, uri: int, props=None):
        if props is None:
            props = {}
        pipeline_syntax = self._create_pipeline_syntax(name, props, uri)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, pipeline_syntax)
            self.pipeline_id = node[0]["node_id"]

    def create_stage_node(self, name: str, parent_context: mlpb.Context, stage_id: int, props=None):
        if props is None:
            props = {}
        parent_id = parent_context.id
        parent_name = parent_context.name
        stage_syntax = self._create_stage_syntax(
            name, props, stage_id, parent_id, parent_name)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, stage_syntax)
            self.stage_id = node[0]["node_id"]
            pc_syntax = self._create_parent_child_syntax(
                "Pipeline", "Stage", self.pipeline_id, self.stage_id, "contains")
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_execution_node(self, name: str, parent_id: int, pipeline_context: mlpb.Context, command: str,
                              execution_id: int,
                              props=None):
        if props is None:
            props = {}
        pipeline_id = pipeline_context.id
        pipeline_name = pipeline_context.name
        execution_syntax = self._create_execution_syntax(
            name, command, props, execution_id, pipeline_id, pipeline_name)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, execution_syntax)
            self.execution_id = node[0]["node_id"]
            pc_syntax = self._create_parent_child_syntax(
                "Stage", "Execution", self.stage_id, self.execution_id, "runs")
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_dataset_node(self, name: str, path: str, uri: str, event: str, execution_id: int,
                            pipeline_context: mlpb.Context,
                            custom_properties=None):
        if custom_properties is None:
            custom_properties = {}
        pipeline_id = pipeline_context.id
        pipeline_name = pipeline_context.name
        dataset_syntax = self._create_dataset_syntax(
            name, path, uri, pipeline_id, pipeline_name, custom_properties)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, dataset_syntax)
            node_id = node[0]["node_id"]
            pc_syntax = self._create_execution_artifacts_link_syntax(
                "Execution", "Dataset", self.execution_id, node_id, event)
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_dataslice_node(self, name: str, path: str, uri: str, parent_name:str,
                            custom_properties=None):
        if custom_properties is None:
            custom_properties = {}
        dataslice_syntax = self._create_dataslice_syntax(
            name, path, uri, custom_properties)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, dataslice_syntax)
            node_id = node[0]["node_id"]
        p_nid = self._get_node("Dataset", parent_name )
        with self.driver.session() as session:
            pc_syntax = self._create_parent_child_syntax(
                "Dataset", "Dataslice", p_nid, node_id, "contains")
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_links(self, source_path:str, target_path:str, relation:str ):
        source_node_id = self._get_node_with_path("Dataset", source_path)
        target_node_id = self._get_node_with_path("Dataslice", target_path)
        with self.driver.session() as session:
            pc_syntax = self._create_parent_child_syntax("Dataset", "Dataslice", source_node_id, target_node_id, relation)
            _ = session.write_transaction(self._run_transaction, pc_syntax)


    def create_model_node(self, name: str, uri: str, event: str, execution_id: str, pipeline_context: mlpb.Context,
                          custom_properties=None):
        if custom_properties is None:
            custom_properties = {}
        pipeline_id = pipeline_context.id
        pipeline_name = pipeline_context.name
        model_syntax = self._create_model_syntax(
            name, uri, pipeline_id, pipeline_name, custom_properties)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, model_syntax)
            node_id = node[0]["node_id"]
            pc_syntax = self._create_execution_artifacts_link_syntax(
                "Execution", "Model", self.execution_id, node_id, event)
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_metrics_node(self, name: str, uri: str, event: str, execution_id: int, pipeline_context: mlpb.Context,
                            custom_properties=None):
        if custom_properties is None:
            custom_properties = {}
        pipeline_id = pipeline_context.id
        pipeline_name = pipeline_context.name
        metrics_syntax = self._create_metrics_syntax(
            name, uri, event, execution_id, pipeline_id, pipeline_name, custom_properties)
        with self.driver.session() as session:
            node = session.write_transaction(
                self._run_transaction, metrics_syntax)
            node_id = node[0]["node_id"]
            pc_syntax = self._create_execution_artifacts_link_syntax(
                "Execution", "Metrics", self.execution_id, node_id, event)
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_artifact_relationships(
            self,
            parent_artifacts,
            child_artifact,
            relation_properties):
        #      f = lambda d: d['Event']
        #      res = {k:list(v) for k,v in groupby(sorted(artifacts, key=f), f)}

        #     for v in res["output"]:

        child_artifact_type = child_artifact["Type"]
        child_artifact_uri = child_artifact["URI"]
        child_name = child_artifact["Name"]
        pipeline_id = child_artifact["Pipeline_Id"]
        for k in parent_artifacts:
            parent_artifact_type = k["Type"]
            parent_artifact_uri = k["URI"]
            parent_name = k["Name"]
            relation = re.sub(
                '\W+', '', re.split(",", k["Execution_Name"])[-1])
            pc_syntax = self._create_parent_child_artifacts_syntax(
                parent_artifact_type,
                child_artifact_type,
                parent_artifact_uri,
                child_artifact_uri,
                parent_name,
                child_name,
                pipeline_id,
                relation,
                relation_properties)
            with self.driver.session() as session:
                _ = session.write_transaction(self._run_transaction, pc_syntax)

    def create_execution_links(
            self,
            parent_artifact_uri,
            parent_artifact_name,
            parent_artifact_type):

        parent_execution_query = "MATCH (n:{}".format(
            parent_artifact_type) + "{uri: '" + parent_artifact_uri + "'}) " \
            "<-[:output]-(f:Execution) Return ID(f)as id, f.uri as uri"

        already_linked_execution_query = "MATCH (f)-[r:linked]->(e2:Execution) " \
            "WHERE r.uri = '{}' RETURN ID(f)as id, f.uri as uri".format(parent_artifact_uri)

        with self.driver.session() as session:
            execution_parent = session.read_transaction(
                self._run_transaction, parent_execution_query)
            executions = {}

            for record in execution_parent:
                p_id = record["id"]
                executions[str(p_id)] = str(record["uri"])

            linked_executions = session.read_transaction(
                self._run_transaction, already_linked_execution_query)
            linked = {}
            for record in linked_executions:
                linked_id = record["id"]
                linked[str(linked_id)] = str(record["uri"])
            unlinked_executions = [
                i for i in executions.keys() if i not in linked.keys()]
            if not unlinked_executions:
                return
            execution_id_to_link = unlinked_executions[0]
            execution_uri = executions[execution_id_to_link]
            pc_syntax = self._create_execution_link_syntax(
                "Execution", "Execution", execution_uri, execution_id_to_link, self.execution_id, "linked", {
                    "Artifact_Name": parent_artifact_name, "uri": parent_artifact_uri})
            _ = session.write_transaction(self._run_transaction, pc_syntax)

    def _get_node(self, node_label: str, node_name: str)->int:
        #Match(n:Metrics) where n.Name contains 'metrics_1' return n
        search_syntax = "MATCH (n:{}) where '{}' in n.Name  \
                              return ID(n) as node_id".format(node_label, node_name)
        print(search_syntax)
        node_id = None
        with self.driver.session() as session:
            nodes = session.read_transaction(
                self._run_transaction, search_syntax)

            node_id = nodes[0]["node_id"]
        return node_id

    def _get_node_with_path(self, node_label: str, node_path: str)->int:
        #Match(n:Metrics) where n.Path contains 'metrics_1' return n
        search_syntax = "MATCH (n:{}) where '{}' in n.Path  \
                              return ID(n) as node_id".format(node_label, node_path)
        print(search_syntax)
        node_id = None
        with self.driver.session() as session:
            nodes = session.read_transaction(
                self._run_transaction, search_syntax)

            node_id = nodes[0]["node_id"]
        return node_id

    @staticmethod
    def _run_transaction(tx, message):
        result = tx.run(message)
        values = []
        for record in result:
            values.append(record)
        return values

    @staticmethod
    def _create_pipeline_syntax(name: str, props: t.Dict, uri: int) -> str:
        props["Name"] = name
        props["uri"] = str(uri)
        props["pipeline_id"] = str(uri)
        props["pipeline_name"] = name
        syntax_str = "MERGE (a:Pipeline {"  # + str(props) + ")"
        for k, v in props.items():
            k = re.sub('\W+', '', k)
            syntax_str = syntax_str + k + ":" + "\"" + v + "\"" + ","
        syntax_str = syntax_str.rstrip(syntax_str[-1])
        syntax_str = syntax_str + "}) RETURN ID(a) as node_id"
        return syntax_str

    # Todo - Verify what is considered as unique node . is it a combination of
    # all properties

    @staticmethod
    def _create_dataset_syntax(name: str, path: str, uri: str, pipeline_id: int, pipeline_name: str,
                               custom_properties):
        custom_properties["Name"] = name
        custom_properties["Path"] = path
        custom_properties["pipeline_id"] = str(pipeline_id)
        custom_properties["pipeline_name"] = pipeline_name
        syntax_str = "MERGE (a:Dataset {uri:\"" + uri + "\"}) SET "
        # props_str = ""
        for k, v in custom_properties.items():
            k = re.sub('\W+', '', k)
            props_str = "a." + k + \
                " = coalesce([x in a." + k + " where x <>\"" + str(v) + "\"], []) + \"" + str(v) + "\","
            syntax_str = syntax_str + props_str
        syntax_str = syntax_str.rstrip(",")
        syntax_str = syntax_str + " RETURN ID(a) as node_id"
        return syntax_str

    @staticmethod
    def _create_dataslice_syntax(name: str, path: str, uri: str,
                               custom_properties):
        custom_properties["Name"] = name
        custom_properties["Path"] = path
        syntax_str = "MERGE (a:Dataslice {uri:\"" + uri + "\"}) SET "
        # props_str = ""
        for k, v in custom_properties.items():
            k = re.sub('\W+', '', k)
            props_str = "a." + k + \
                " = coalesce([x in a." + k + " where x <>\"" + str(v) + "\"], []) + \"" + str(v) + "\","
            syntax_str = syntax_str + props_str
        syntax_str = syntax_str.rstrip(",")
        syntax_str = syntax_str + " RETURN ID(a) as node_id"
        return syntax_str

    @staticmethod
    def _create_model_syntax(name: str, uri: str, pipeline_id: int, pipeline_name: str, custom_properties):
        custom_properties["Name"] = name
        custom_properties["uri"] = uri
        custom_properties["pipeline_id"] = str(pipeline_id)
        custom_properties["pipeline_name"] = pipeline_name
        syntax_str = "MERGE (a:Model {"  # + str(props) + ")"
        for k, v in custom_properties.items():
            k = re.sub('\W+', '', k)
            syntax_str = syntax_str + k + ":" + "\"" + str(v) + "\"" + ","
        syntax_str = syntax_str.rstrip(syntax_str[-1])
        syntax_str = syntax_str + "})"
        syntax_str = syntax_str + " RETURN ID(a) as node_id"
        return syntax_str

    @staticmethod
    def _create_metrics_syntax(name: str, uri: str, event: str, execution_id: int, pipeline_id: int,
                               pipeline_name: str, custom_properties):
        custom_properties["Name"] = name
        custom_properties["uri"] = uri
        # custom_properties["execution_id"] = str(execution_id)
        custom_properties["pipeline_id"] = str(pipeline_id)
        custom_properties["pipeline_name"] = pipeline_name
        syntax_str = "MERGE (a:Metrics {"  # + str(props) + ")"
        for k, v in custom_properties.items():
            k = re.sub('\W+', '', k)
            syntax_str = syntax_str + k + ":" + "\"" + str(v) + "\"" + ","
        syntax_str = syntax_str.rstrip(syntax_str[-1])
        syntax_str = syntax_str + "})"
        syntax_str = syntax_str + " RETURN ID(a) as node_id"
        return syntax_str

    @staticmethod
    def _create_stage_syntax(name: str, props: t.Dict, uri: int, pipeline_id: int, pipeline_name: str) -> str:
        props["Name"] = name
        props["uri"] = str(uri)
        props["pipeline_id"] = str(pipeline_id)
        props["pipeline_name"] = pipeline_name
        syntax_str = "MERGE (a:Stage {"  # + str(props) + ")"
        for k, v in props.items():
            k = re.sub('\W+', '', k)
            syntax_str = syntax_str + k + ":" + "\"" + str(v) + "\"" + ","

        syntax_str = syntax_str.rstrip(syntax_str[-1])
        syntax_str = syntax_str + "}) RETURN ID(a) as node_id"
        return syntax_str


    @staticmethod
    def _create_parent_child_syntax(parent_label: str, child_label: str, parent_id: int, child_id: int, relation: str):
        parent_child_syntax = "MATCH (a:{}), (b:{}) where ID(a) = {} AND ID(b) = {} MERGE (a)-[r:{}]->(b) \
                              return type(r)".format(parent_label, child_label, parent_id, child_id, relation)
        return parent_child_syntax

    @staticmethod
    def _create_execution_artifacts_link_syntax(parent_label: str, child_label: str, parent_id: int, child_id: int,
                                                relation: str):
        if relation.lower() == "input":
            parent_child_syntax = "MATCH (a:{}), (b:{}) where ID(a) = {} AND ID(b) = {} MERGE (a)<-[r:{}]-(b) \
                              return type(r)".format(parent_label, child_label, parent_id, child_id, relation)
        else:
            parent_child_syntax = "MATCH (a:{}), (b:{}) where ID(a) = {} AND ID(b) = {} MERGE (a)-[r:{}]->(b) \
                              return type(r)".format(parent_label, child_label, parent_id, child_id, relation)

        return parent_child_syntax

    @staticmethod
    def _create_execution_link_syntax(parent_label: str, child_label: str, parent_uri: str, parent_id: str,
                                      child_id: int,
                                      relation: str, relation_properties: t.Dict):
        """
        MATCH
        (a:Person),
        (b:Person)
        WHERE a.name = 'A' AND b.name = 'B'
        CREATE (a)-[r:RELTYPE]->(b)
        RETURN type(r)
        """
        parent_child_syntax_1 = "MATCH (a:{}), (b:{}) WHERE a.uri = '{}' AND ID(a) = {} AND  ID(b) = {} ".format(
            parent_label, child_label, parent_uri, parent_id, child_id)
        parent_child_syntax_2 = "MERGE (a)-[r:{}".format(relation)
        parent_child_syntax_3 = "{"
        for k, v in relation_properties.items():
            parent_child_syntax_3 = parent_child_syntax_3 + k + ":" + "\"" + v + "\"" + ","
        parent_child_syntax_3 = parent_child_syntax_3.rstrip(
            parent_child_syntax_3[-1])
        parent_child_syntax_4 = "}]->(b) RETURN type(r)"
        parent_child_syntax = parent_child_syntax_1 + parent_child_syntax_2 \
            + parent_child_syntax_3 + parent_child_syntax_4
        return parent_child_syntax

    @staticmethod
    def _create_parent_child_artifacts_syntax(parent_label: str, child_label: str, parent_uri: str, child_uri: str,
                                              parent_name: str, child_name: str, pipeline_id: int, relation: str,
                                              relation_properties: t.Dict):
        """
        MATCH
        (a:Person),
        (b:Person)
        WHERE a.name = 'A' AND b.name = 'B'
        CREATE (a)-[r:RELTYPE]->(b)
        RETURN type(r)
        """
        parent_child_syntax_1 = "MATCH (a:{}), (b:{}) WHERE a.uri = '{}' AND b.uri = '{}' ".format(
            parent_label, child_label, parent_uri, child_uri)
        parent_child_syntax_2 = "MERGE (a)-[r:{}".format(relation)
        parent_child_syntax_3 = "{"
        for k, v in relation_properties.items():
            parent_child_syntax_3 = parent_child_syntax_3 + \
                k + ":" + "\"" + str(v) + "\"" + ","
        parent_child_syntax_3 = parent_child_syntax_3.rstrip(
            parent_child_syntax_3[-1])
        parent_child_syntax_4 = "}]->(b) RETURN type(r)"
        parent_child_syntax = parent_child_syntax_1 + parent_child_syntax_2 + \
            parent_child_syntax_3 + parent_child_syntax_4
        return parent_child_syntax

    @staticmethod
    def _create_execution_syntax(name: str, command: str, props: t.Dict, uri: int, pipeline_id: int,
                                 pipeline_name: str) -> str:
        props["Name"] = name
        props["Command"] = command
        props["uri"] = str(uri)
        props["pipeline_id"] = str(pipeline_id)
        props["pipeline_name"] = pipeline_name
        syntax_str = "MERGE (a:Execution {"  # + str(props) + ")"
        for k, v in props.items():
            k = re.sub('\W+', '', k)
            v = str(v)
            syntax_str = syntax_str + k + ":" + "\"" + v + "\"" + ","

        syntax_str = syntax_str.rstrip(syntax_str[-1])
        syntax_str = syntax_str + "}) RETURN ID(a) as node_id"
        return syntax_str



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import configparser
import os
import base64

class CmfConfig:
    @staticmethod
    def write_config(config_file:str, section_name:str, dict_of_attr, file_exists=False):
        config = configparser.ConfigParser()
        if file_exists:
            with open(config_file, "r") as file:
                config.read_file(file)
            if config.has_section(section_name):
                config.remove_section(section_name)
        config.add_section(section_name)
        for key, value in dict_of_attr.items():
            if section_name =='neo4j' and key == 'password':
                value = str(base64.b64encode(value.encode("UTF-8")), "UTF-8")
            config.set(section_name, key, value)
        with open(config_file, "w") as file:
            config.write(file)

    @staticmethod
    def read_config(config_file:str):
        output_dict = {}
        config_data = configparser.ConfigParser()
        with open(config_file, "r") as file:
            config_data.read_file(file)
        sections = config_data.sections()
        for sec in sections:
            sec_data = config_data[sec]
            for key in sec_data:
                value = sec_data.get(key)
                if sec == 'neo4j' and key == 'password':
                    encoded_pass = bytes(value,"utf-8")
                    value = base64.b64decode(encoded_pass).decode("utf-8")
                key = f"{sec}-{key}"
                output_dict[key] = value
        return output_dict



import itertools
import re
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout
import pandas as pd
from cmflib import cmfquery
import dvc
import json
import random
import warnings

warnings.filterwarnings("ignore")
def query_visualization_ArtifactExecution(mlmd_path, pipeline_name):
    file_path="/cmf-server/data/static/data.json"
    query = cmfquery.CmfQuery(mlmd_path)
    stages = query.get_pipeline_stages(pipeline_name)
    node_id_name_list=[]
    link_src_trgt_list=[]
    seen_ids_and_names = set()
    for stage in stages:
        executions = query.get_all_executions_in_stage(stage)
        for i in executions.id.to_list():
            artifacts = query.get_all_artifacts_for_execution(i)
#            print(artifacts['name'],type(artifacts))
            print("___________________")
            for i in range(len(artifacts)):
                node_id_name={}
                link_src_trgt_={}
                node_id_name["id"]=artifacts.loc[i, "id"]
                node_id_name["name"]=artifacts.loc[i, "name"]
                id_and_name = (node_id_name['id'], node_id_name['name'])
                if id_and_name not in seen_ids_and_names:
                    seen_ids_and_names.add(id_and_name)
                    node_id_name_list.append(node_id_name)
                if artifacts.loc[i, "event"] == "INPUT":
                    link_src_trgt["source"]=stage
                    link_src_trgt["target"]=artifacts.loc[i,"id"]
    for stage in stages:
        id_list=[100:200]
        node_id_name["id"]
#            for i in artifacts:
#                print(i['name'])
        print(stage)
        
query_visualization_ArtifactExecution("/home/chobey/cmf-server/data/mlmd","image")



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os
import subprocess
import time

from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cli.utils import check_minio_server
from cmflib.utils.helper_functions import generate_osdf_token
from cmflib.utils.helper_functions import is_url
from cmflib.utils.dvc_config import DvcConfig
from cmflib.dvc_wrapper import dvc_push
from cmflib.dvc_wrapper import dvc_add_attribute
from cmflib.utils.cmf_config import CmfConfig

class CmdArtifactPush(CmdBase):
    def run(self):
        result = ""
        dvc_config_op = DvcConfig.get_dvc_config()
        cmf_config_file = os.environ.get("CONFIG_FILE", ".cmfconfig")
        cmf_config={}
        cmf_config=CmfConfig.read_config(cmf_config_file)
        out_msg = check_minio_server(dvc_config_op)
        if dvc_config_op["core.remote"] == "minio" and out_msg != "SUCCESS":
            return "MinioS3 server failed to start!!!"
        if dvc_config_op["core.remote"] == "osdf":
            #print("key_id="+cmf_config["osdf-key_id"])
            dynamic_password = generate_osdf_token(cmf_config["osdf-key_id"],cmf_config["osdf-key_path"],cmf_config["osdf-key_issuer"])
            #print("Dynamic Password"+dynamic_password)
            dvc_add_attribute(dvc_config_op["core.remote"],"password",dynamic_password)
            #The Push URL will be something like: https://<Path>/files/md5/[First Two of MD5 Hash]
            result = dvc_push()
            #print(result)
            return result

        current_directory = os.getcwd()
        # Default path of mlmd file
        mlmd_file_name = "./mlmd"
        if self.args.file_name:
            mlmd_file_name = self.args.file_name
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
            current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):
            return f"ERROR: {mlmd_file_name} doesn't exists in {current_directory} directory."

        # creating cmfquery object
        query = cmfquery.CmfQuery(mlmd_file_name)

         # Put a check to see whether pipline exists or not
        pipeline_name = self.args.pipeline_name
        if not query.get_pipeline_id(pipeline_name) > 0:
            return f"ERROR: Pipeline {pipeline_name} doesn't exist!!"

        stages = query.get_pipeline_stages(self.args.pipeline_name)
        executions = []
        identifiers = []

        for stage in stages:
            # getting all executions for stages
            executions = query.get_all_executions_in_stage(stage)
            # check if stage has executions
            if len(executions) > 0:
                # converting it to dictionary
                dict_executions = executions.to_dict("dict")
                for id in dict_executions["id"].values():
                    identifiers.append(id)
            else:
                print("No Executions found for " + stage + " stage.")

        names = []
        if len(identifiers) == 0:  # check if there are no executions
            return "No executions found."
        for identifier in identifiers:
            artifacts = query.get_all_artifacts_for_execution(
                 identifier
            )  # getting all artifacts with id
            # dropping artifact with type 'metrics' as metrics doesn't have physical file
            artifacts = artifacts[artifacts['type'] != 'Metrics']
            # adding .dvc at the end of every file as it is needed for pull
            artifacts['name'] = artifacts['name'].apply(lambda name: f"{name.split(':')[0]}.dvc")
            names.extend(artifacts['name'].tolist())
        file_set = set(names)
        result = dvc_push(list(file_set))
        return result
      
def add_parser(subparsers, parent_parser):
    HELP = "Push artifacts to the user configured artifact repo."

    parser = subparsers.add_parser(
        "push",
        parents=[parent_parser],
        description="Push artifacts to the user configured artifact repo.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", "--file_name", help="Specify mlmd file name.", metavar="<file_name>"
    )

    parser.set_defaults(func=CmdArtifactPush)



import time
from cmflib import cmfquery
from collections import deque, defaultdict
#from get_data import get_all_artifact_ids, get_all_exe_ids
from typing import List, Dict, Any

async def query_artifact_lineage_d3tree(mlmd_path: str, pipeline_name: str, dict_of_art_ids: Dict) -> List[List[Dict[str, Any]]]:
    query = cmfquery.CmfQuery(mlmd_path)
    id_name = {}
    child_parent_artifact_id = {}
    for type_, df in dict_of_art_ids[pipeline_name].items():
        for index, row in df.iterrows():
            #creating a dictionary of id and artifact name {id:artifact name}
            artifact_id = row['id']  # This will be an integer
            id_name[artifact_id] = modify_arti_name(row["name"], type_)
            one_hop_parent_artifacts = query.get_one_hop_parent_artifacts_with_id(artifact_id)  # get immediate artifacts     
            child_parent_artifact_id[artifact_id] = []      # assign empty dict for artifact with no parent artifact
            if not one_hop_parent_artifacts.empty:        # if artifact have parent artifacts             
                child_parent_artifact_id[artifact_id] = list(one_hop_parent_artifacts["id"])
    data_organized = topological_sort(child_parent_artifact_id, id_name)
    return data_organized

def topological_sort(input_data, artifact_name_id_dict) -> List[Dict]:
    # Initialize in-degree of all nodes to 0
    in_degree = {node: 0 for node in input_data}
    # Initialize adjacency list
    adj_list = defaultdict(list)

    # Fill the adjacency list and in-degree dictionary
    for node, dependencies in input_data.items():
        for dep in dependencies:
            adj_list[dep].append(node)
            in_degree[node] += 1

    # Queue for nodes with in-degree 0
    zero_in_degree_queue = deque([node for node, degree in in_degree.items() if degree == 0])
    topo_sorted_nodes = []

    while zero_in_degree_queue:
        current_node = zero_in_degree_queue.popleft()
        topo_sorted_nodes.append(current_node)
        for neighbor in adj_list[current_node]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                zero_in_degree_queue.append(neighbor)
    # Transform sorted nodes into the required output format
    parent_dict = defaultdict(list)
    for id_val in topo_sorted_nodes:   # topo_sorted_nodes = ['1','2','3','4']
        if id_val in input_data:       # input_data = {"child_id":[parents_id]}, for example {"4":['3','7','9']}
            parents = tuple(sorted(input_data[id_val]))
            parent_dict[parents].append({'id': artifact_name_id_dict[id_val],'parents': [artifact_name_id_dict[parent] for parent in input_data[id_val]]})
    output_data= list(parent_dict.values()) 
    return output_data

def modify_arti_name(arti_name, type):
    # artifact_name optimization based on artifact type.["Dataset","Model","Metrics"]
    try:
        name = ""
        if type == "Metrics" :   # Example metrics:4ebdc980-1e7c-11ef-b54c-25834a9c665c:388 -> metrics:4ebd:388
            name = f"{arti_name.split(':')[0]}:{arti_name.split(':')[1][:4]}:{arti_name.split(':')[2]}"
        elif type == "Model":
            #first split on ':' then on '/' to get name. Example 'Test-env/prepare:uuid:32' -> prepare_uuid
            name = arti_name.split(':')[-3].split("/")[-1] + ":" + arti_name.split(':')[-2][:4]
        elif type == "Dataset":
            # Example artifacts/data.xml.gz:236d9502e0283d91f689d7038b8508a2 -> data.xml.gz 
            name = arti_name.split(':')[-2] .split("/")[-1]  
        elif type == "Dataslice":
            # cmf_artifacts/dataslices/ecd6dcde-4f3b-11ef-b8cd-f71a4cc9ba38/slice-1:e77e3466872898fcf2fa22a3752bc1ca
            dataslice_part1 = arti_name.split("/",1)[1] #remove cmf_artifacts/
            # dataslices/ecd6dcde-4f3b-11ef-b8cd-f71a4cc9ba38/slice-1 + : + e77e
            name = dataslice_part1.rsplit(":",-1)[0] + ":" + dataslice_part1.rsplit(":",-1)[-1][:4]
        elif type == "Step_Metrics":
            #cmf_artifacts/metrics/1a86b01c-4da9-11ef-b8cd-f71a4cc9ba38/training_metrics:d7c32a3f4fce4888c905de07ba253b6e:3:2029c720-4da9-11ef-b8cd-f71a4cc9ba38
            step_new = arti_name.split("/",1)[1]     #remove cmf_artifacts/
            step_metrics_part2 = arti_name.rsplit(":")
            # metrics/1a86b01c-4da9-11ef-b8cd-f71a4cc9ba38/training_metrics: + d7c3 + : +3 + : + 2029
            name = step_new.rsplit(":",-3)[0] + ":" + step_metrics_part2[-3][:4] + ":" + step_metrics_part2[-2] + ":" + step_metrics_part2[-1][:4]
        else:
            name = arti_name  
    except Exception as e:
        print(f"Error parsing artifact name: {e}")
        name = arti_name  # Fallback to the original arti_name in case of error
    return name






###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
import abc
import json
import logging
import typing as t
from enum import Enum
from google.protobuf.json_format import MessageToDict

import pandas as pd
from ml_metadata.metadata_store import metadata_store
from ml_metadata.proto import metadata_store_pb2 as mlpb

from cmflib.mlmd_objects import CONTEXT_LIST

__all__ = ["CmfQuery"]

logger = logging.getLogger(__name__)


class _KeyMapper(abc.ABC):
    """Map one key (string) to another key (string) using a predefined strategy.
    Args:
        on_collision: What to do if the mapped key already exists in the target dictionary.
    """

    class OnCollision(Enum):
        """What to do when a mapped key exists in the target dictionary."""

        DO_NOTHING = 0
        """Ignore the collision and overwrite the value associated with this key."""
        RESOLVE = 1
        """Resolve it by appending `_INDEX` where INDEX as the smallest positive integer that avoids this collision."""
        RAISE_ERROR = 2
        """Raise an exception."""

    def __init__(self, on_collision: OnCollision = OnCollision.DO_NOTHING) -> None:
        self.on_collision = on_collision

    def get(self, d: t.Mapping, key: t.Any) -> t.Any:
        """Return new (mapped) key.
        Args:
            d: Dictionary to update with the mapped key.
            key: Source key name.
         Returns:
            A mapped (target) key to be used with the `d` dictionary.
        """
        new_key = self._get(key)
        if new_key in d:
            if self.on_collision == _KeyMapper.OnCollision.RAISE_ERROR:
                raise ValueError(f"Mapped key ({key} -> {new_key}) already exists.")
            elif self.on_collision == _KeyMapper.OnCollision.RESOLVE:
                _base_key, index = new_key, 0
                while new_key in d:
                    index += 1
                    new_key = f"{_base_key}_{index}"
        return new_key

    @abc.abstractmethod
    def _get(self, key: t.Any) -> t.Any:
        """Mapp a source key to a target key.
        Args:
            key: Source key.
        Returns:
            Target key.
        """
        raise NotImplementedError()


class _DictMapper(_KeyMapper):
    """Use dictionaries to specify key mappings (source -> target)."""

    def __init__(self, mappings: t.Mapping, **kwargs) -> None:
        super().__init__(**kwargs)
        self.mappings = mappings

    def _get(self, key: t.Any) -> t.Any:
        return self.mappings.get(key, key)


class _PrefixMapper(_KeyMapper):
    """Prepend a constant prefix to produce a mapped key."""

    def __init__(self, prefix: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.prefix = prefix

    def _get(self, key: t.Any) -> t.Any:
        return self.prefix + key


class CmfQuery(object):
    """CMF Query communicates with the MLMD database and implements basic search and retrieval functionality.

    This class has been designed to work with the CMF framework. CMF alters names of pipelines, stages and artifacts
    in various ways. This means that actual names in the MLMD database will be different from those originally provided
    by users via CMF API. When methods in this class accept `name` parameters, it is expected that values of these
    parameters are fully-qualified names of respective entities.


    Args:
        filepath: Path to the MLMD database file.
    """

    def __init__(self, filepath: str = "mlmd") -> None:
        config = mlpb.ConnectionConfig()
        config.sqlite.filename_uri = filepath
        self.store = metadata_store.MetadataStore(config)

    @staticmethod
    def _copy(
        source: t.Mapping, target: t.Optional[t.Dict] = None, key_mapper: t.Optional[t.Union[t.Dict, _KeyMapper]] = None
    ) -> t.Dict:
        """Create copy of `source` and return it, reuse `target` if not None.

        Args:
            source: Input dict-like object to create copy.
            target: If not None, this will be reused and returned. If None, new dict will be created.
            key_mapper: Dictionary containing how to map keys in `source`, e.g., {"key_in": "key_out"} means the
                        key in `source` named "key_in" should be renamed to "key_out" in output dictionary object, or
                        instance of _KeyMapper.
        Returns:
            If `target` is not None, it is returned containing data from `source`. Else, new object is returned.
        """
        if target is None:
            target = {}
        if key_mapper is None:
            key_mapper = _DictMapper({})
        elif isinstance(key_mapper, dict):
            key_mapper = _DictMapper(key_mapper)
        assert isinstance(key_mapper, _KeyMapper), f"Invalid key_mapper type (type={type(key_mapper)})."

        for key, value in source.items():
            if value.HasField("string_value"):
                value = value.string_value
            elif value.HasField("int_value"):
                value = value.int_value
            else:
                value = value.double_value

            target[key_mapper.get(target, key)] = value

        return target

    @staticmethod
    def _transform_to_dataframe(
        node: t.Union[mlpb.Execution, mlpb.Artifact], d: t.Optional[t.Dict] = None
    ) -> pd.DataFrame:
        """Transform MLMD entity `node` to pandas data frame.

        Args:
            node: MLMD entity to transform.
            d: Pre-populated dictionary of KV-pairs to associate  with `node` (will become columns in output table).
        Returns:
            Pandas data frame with one row containing data from `node`.
        """
        if d is None:
            d = {}

        d = CmfQuery._copy(
            source=node.properties,
            target=d#, # renaming properties with prefix properties has impact in server GUI 
            #key_mapper=_PrefixMapper("properties_", on_collision=_KeyMapper.OnCollision.RESOLVE),
        )
        d = CmfQuery._copy(
            source=node.custom_properties,
            target=d, # renaming custom_properties with prefix custom_properties has impact in server GUI 
            key_mapper=_PrefixMapper("custom_properties_", on_collision=_KeyMapper.OnCollision.RESOLVE),
        )

        return pd.DataFrame(
            d,
            index=[
                0,
            ],
        )

    @staticmethod
    def _as_pandas_df(elements: t.Iterable, transform_fn: t.Callable[[t.Any], pd.DataFrame]) -> pd.DataFrame:
        """Convert elements in `elements` to rows in pandas data frame using `transform_fn` function.

        Args:
            elements: Collection with items to be converted to tabular representation, each item becomes one row.
            transform_fn: A callable object that takes one element in `elements` and returns its tabular representation
                (pandas data frame with one row).
        Returns:
            Pandas data frame containing representation of elements in `elements` with one row being one element.
        """
        df = pd.DataFrame()
        for element in elements:
            df = pd.concat([df, transform_fn(element)], sort=True, ignore_index=True)
        return df

    def _get_pipelines(self, name: t.Optional[str] = None) -> t.List[mlpb.Context]:
        pipelines: t.List[mlpb.Context] = self.store.get_contexts_by_type("Parent_Context")
        """Return list of pipelines with the given name.

        Args:
            name: Piepline name or None to return all pipelines.
        Returns:
            List of objects associated with pipelines.
        """
        if name is not None:
            pipelines = [pipeline for pipeline in pipelines if pipeline.name == name]
        return pipelines

    def _get_pipeline(self, name: str) -> t.Optional[mlpb.Context]:
        """Return a pipeline with the given name or None if one does not exist.
        Args:
            name: Pipeline name.
        Returns:
            A pipeline object if found, else None.
        """
        pipelines: t.List = self._get_pipelines(name)
        if pipelines:
            if len(pipelines) >= 2:
                logger.debug("Found %d pipelines with '%s' name.", len(pipelines), name)
            return pipelines[0]
        return None

    def _get_stages(self, pipeline_id: int) -> t.List[mlpb.Context]:
        """Return stages for the given pipeline.

        Args:
            pipeline_id: Pipeline ID.
        Returns:
            List of associated pipeline stages.
        """
        return self.store.get_children_contexts_by_context(pipeline_id)

    def _get_executions(self, stage_id: int, execution_id: t.Optional[int] = None) -> t.List[mlpb.Execution]:
        """Return executions of the given stage.

        Args:
            stage_id: Stage identifier.
            execution_id: If not None, return only execution with this ID.
        Returns:
            List of executions matching input parameters.
        """
        executions: t.List[mlpb.Execution] = self.store.get_executions_by_context(stage_id)
        if execution_id is not None:
            executions = [execution for execution in executions if execution.id == execution_id]
        return executions

    def _get_executions_by_input_artifact_id(self, artifact_id: int,pipeline_id: str = None) -> t.List[int]:
        """Return stage executions that consumed given input artifact.

        Args:
            artifact_id: Identifier of the input artifact.
        Returns:
            List of stage executions that consumed the given artifact.
        """
        execution_ids = list(set(
            event.execution_id
            for event in self.store.get_events_by_artifact_ids([artifact_id])
            if event.type == mlpb.Event.INPUT
        ))
        
        if pipeline_id != None:
            list_exec=self.store.get_executions_by_id(execution_ids)
            execution_ids=[]
            for exe in list_exec:
                if (self._transform_to_dataframe(exe).Pipeline_id.to_string(index=False)) == str(pipeline_id):
                    execution_ids.append(exe.id)
        return execution_ids

    def _get_executions_by_output_artifact_id(self, artifact_id: int, pipeline_id: str = None) -> t.List[int]:
        """Return stage execution that produced given output artifact.

        Args:
            artifact_id: Identifier of the output artifact.
        Return:
            List of stage executions, should probably be size of 1 or zero.
        """
        execution_ids: t.List[int] = [
            event.execution_id
            for event in self.store.get_events_by_artifact_ids([artifact_id])
            if event.type == mlpb.Event.OUTPUT
        ]
        # According to CMF, it's OK to have multiple executions that produce the same exact artifact.
        # if len(execution_ids) >= 2:
        #     logger.warning("%d executions claim artifact (id=%d) as output.", len(execution_ids), artifact_id)
        if pipeline_id != None:
            list_exec = self.store.get_executions_by_id(execution_ids)
            execution_ids = []
            for exe in list_exec:
                if (self._transform_to_dataframe(exe).Pipeline_id.to_string(index=False)) == str(pipeline_id):
                    execution_ids.append(exe.id)
        return execution_ids

    def _get_artifact(self, name: str) -> t.Optional[mlpb.Artifact]:
        """Return artifact with the given name or None.
        Args:
            name: Fully-qualified name (e.g., artifact hash is added to the name), so name collisions across different
                  artifact types are not issues here.
        Returns:
            Artifact or None (if not found).
        """
        name = name.strip()
        for artifact in self.store.get_artifacts():
            if artifact.name == name:
                return artifact
        return None

    def _get_output_artifacts(self, execution_ids: t.List[int]) -> t.List[int]:
        """Return output artifacts for the given executions.

        Artifacts are uniquely identified by their hashes in CMF, and so, when executions produce the same exact file,
        they will claim this artifact as an output artifact, and so same artifact can have multiple producer
        executions.

        Args:
            execution_ids: List of execution identifiers to return output artifacts for.
        Returns:
            List of output artifact identifiers.
        """
        artifact_ids: t.List[int] = [
            event.artifact_id
            for event in self.store.get_events_by_execution_ids(set(execution_ids))
            if event.type == mlpb.Event.OUTPUT
        ]
        unique_artifact_ids = set(artifact_ids)
        if len(unique_artifact_ids) != len(artifact_ids):
            logger.warning("Multiple executions claim the same output artifacts")
#        artifacts=self.get_all_artifacts_by_ids_list(list(unique_artifact_ids))
#        for key,val in artifacts.iterrows():
#                print(val["name"])
        return list(unique_artifact_ids)

    def _get_input_artifacts(self, execution_ids: t.List[int]) -> t.List[int]:
        """Return input artifacts for the given executions.

        Args:
            execution_ids: List of execution identifiers to return input artifacts for.
        Returns:
            List of input artifact identifiers.
        """
        artifact_ids = set(
            event.artifact_id
            for event in self.store.get_events_by_execution_ids(set(execution_ids))
            if event.type == mlpb.Event.INPUT
        )
        return list(artifact_ids)

    def get_pipeline_names(self) -> t.List[str]:
        """Return names of all pipelines.

        Returns:
            List of all pipeline names.
        """
        return [ctx.name for ctx in self._get_pipelines()]

    def get_pipeline_id(self, pipeline_name: str) -> int:
        """Return pipeline identifier for the pipeline names `pipeline_name`.
        Args:
            pipeline_name: Name of the pipeline.
        Returns:
            Pipeline identifier or -1 if one does not exist.
        """
        pipeline: t.Optional[mlpb.Context] = self._get_pipeline(pipeline_name)
        return -1 if not pipeline else pipeline.id

    def get_pipeline_stages(self, pipeline_name: str) -> t.List[str]:
        """Return list of pipeline stages for the pipeline with the given name.

        Args:
            pipeline_name: Name of the pipeline for which stages need to be returned. In CMF, there are no different
                pipelines with the same name.
        Returns:
            List of stage names associated with the given pipeline.
        """
        stages = []
        for pipeline in self._get_pipelines(pipeline_name):
            stages.extend(stage.name for stage in self._get_stages(pipeline.id))
        return stages

    def get_all_exe_in_stage(self, stage_name: str) -> t.List[mlpb.Execution]:
        """Return list of all executions for the stage with the given name.

        Args:
            stage_name: Name of the stage. Before stages are recorded in MLMD, they are modified (e.g., pipeline name
                        will become part of the stage name). So stage names from different pipelines will not collide.
        Returns:
            List of executions for the given stage.
        """
        for pipeline in self._get_pipelines():
            for stage in self._get_stages(pipeline.id):
                if stage.name == stage_name:
                    return self.store.get_executions_by_context(stage.id)
        return []

    def get_all_executions_by_ids_list(self, exe_ids: t.List[int]) -> pd.DataFrame:
        """Return executions for given execution ids list as a pandas data frame.

        Args:
            exe_ids: List of execution identifiers.

        Returns:
            Data frame with all executions for the list of given execution identifiers.
        """

        df = pd.DataFrame()
        executions = self.store.get_executions_by_id(exe_ids)
        for exe in executions:
            d1 = self._transform_to_dataframe(exe)
            df = pd.concat([df, d1], sort=True, ignore_index=True)
        return df

    def get_all_artifacts_by_context(self, pipeline_name: str) -> pd.DataFrame:
        """Return artifacts for given pipeline name as a pandas data frame.

        Args:
            pipeline_name: Name of the pipeline.

        Returns:
            Data frame with all artifacts associated with given pipeline name.
        """
        df = pd.DataFrame()
        contexts = self.store.get_contexts_by_type("Parent_Context")
        context_id = self.get_pipeline_id(pipeline_name)
        for ctx in contexts:
            if ctx.id == context_id:
                child_contexts = self.store.get_children_contexts_by_context(ctx.id)
                for cc in child_contexts:
                    artifacts = self.store.get_artifacts_by_context(cc.id)
                    for art in artifacts:
                        d1 = self.get_artifact_df(art)
                        df = pd.concat([df, d1], sort=True, ignore_index=True)
        return df

    def get_all_artifacts_by_ids_list(self, artifact_ids: t.List[int]) -> pd.DataFrame:
        """Return all artifacts for the given artifact ids list.

        Args:
            artifact_ids: List of artifact identifiers

        Returns:
            Data frame with all artifacts for the given artifact ids list.
        """
        df = pd.DataFrame()
        artifacts = self.store.get_artifacts_by_id(artifact_ids)
        for art in artifacts:
            d1 = self.get_artifact_df(art)
            df = pd.concat([df, d1], sort=True, ignore_index=True)
        return df

    def get_all_executions_in_stage(self, stage_name: str) -> pd.DataFrame:
        """Return executions of the given stage as pandas data frame.
        Args:
            stage_name: Stage name. See doc strings for the prev method.
        Returns:
            Data frame with all executions associated with the given stage.
        """
        df = pd.DataFrame()
        for pipeline in self._get_pipelines():
            for stage in self._get_stages(pipeline.id):
                if stage.name == stage_name:
                    for execution in self._get_executions(stage.id):
                        ex_as_df: pd.DataFrame = self._transform_to_dataframe(
                            execution, {"id": execution.id, "name": execution.name}
                        )
                        df = pd.concat([df, ex_as_df], sort=True, ignore_index=True)
        return df

    def get_artifact_df(self, artifact: mlpb.Artifact, d: t.Optional[t.Dict] = None) -> pd.DataFrame:
        """Return artifact's data frame representation.

        Args:
            artifact: MLMD entity representing artifact.
            d: Optional initial content for data frame.
        Returns:
            A data frame with the single row containing attributes of this artifact.
        """
        if d is None:
            d = {}
        d.update(
            {
                "id": artifact.id,
                "type": self.store.get_artifact_types_by_id([artifact.type_id])[0].name,
                "uri": artifact.uri,
                "name": artifact.name,
                "create_time_since_epoch": artifact.create_time_since_epoch,
                "last_update_time_since_epoch": artifact.last_update_time_since_epoch,
            }
        )
        return self._transform_to_dataframe(artifact, d)

    def get_all_artifacts(self) -> t.List[str]:
        """Return names of all artifacts.

        Returns:
            List of all artifact names.
        """
        return [artifact.name for artifact in self.store.get_artifacts()]

    get_artifact_names = get_all_artifacts

    def get_artifact(self, name: str) -> t.Optional[pd.DataFrame]:
        """Return artifact's data frame representation using artifact name.

        Args:
            name: Artifact name.
        Returns:
            Pandas data frame with one row containing attributes of this artifact.
        """
        artifact: t.Optional[mlpb.Artifact] = self._get_artifact(name)
        if artifact:
            return self.get_artifact_df(artifact)
        return None

    def get_all_artifacts_for_execution(self, execution_id: int) -> pd.DataFrame:
        """Return input and output artifacts for the given execution.

        Args:
            execution_id: Execution identifier.
        Return:
            Data frame containing input and output artifacts for the given execution, one artifact per row.
        """
        df = pd.DataFrame()
        for event in self.store.get_events_by_execution_ids([execution_id]):
            event_type = "INPUT" if event.type == mlpb.Event.Type.INPUT else "OUTPUT"
            for artifact in self.store.get_artifacts_by_id([event.artifact_id]):
                df = pd.concat(
                    [df, self.get_artifact_df(artifact, {"event": event_type})], sort=True, ignore_index=True
                )
        return df

    def get_all_artifact_types(self) -> t.List[str]:
        """Return names of all artifact types.

        Returns:
            List of all artifact types.
        """
        artifact_list = self.store.get_artifact_types()
        types=[i.name for i in artifact_list]
        return types

    def get_all_executions_for_artifact(self, artifact_name: str) -> pd.DataFrame:
        """Return executions that consumed and produced given artifact.

        Args:
            artifact_name: Artifact name.
        Returns:
            Pandas data frame containing stage executions, one execution per row.
        """
        df = pd.DataFrame()

        artifact: t.Optional = self._get_artifact(artifact_name)
        if not artifact:
            return df

        for event in self.store.get_events_by_artifact_ids([artifact.id]):
            stage_ctx = self.store.get_contexts_by_execution(event.execution_id)[0]
            linked_execution = {
                "Type": "INPUT" if event.type == mlpb.Event.Type.INPUT else "OUTPUT",
                "execution_id": event.execution_id,
                "execution_name": self.store.get_executions_by_id([event.execution_id])[0].name,
                "execution_type_name":self.store.get_executions_by_id([event.execution_id])[0].properties['Execution_type_name'],
                "stage": stage_ctx.name,
                "pipeline": self.store.get_parent_contexts_by_context(stage_ctx.id)[0].name,
            }
            d1 = pd.DataFrame(
                linked_execution,
                index=[
                    0,
                ],
            )
            df = pd.concat([df, d1], sort=True, ignore_index=True)
        return df

    def get_one_hop_child_artifacts(self, artifact_name: str, pipeline_id: str = None) -> pd.DataFrame:
        """Get artifacts produced by executions that consume given artifact.

        Args:
            artifact name: Name of an artifact.
        Return:
            Output artifacts of all executions that consumed given artifact.
        """
        artifact: t.Optional = self._get_artifact(artifact_name)
        if not artifact:
            return pd.DataFrame()

        # Get output artifacts of executions consumed the above artifact.
        artifacts_ids = self._get_output_artifacts(self._get_executions_by_input_artifact_id(artifact.id,pipeline_id))
        return self._as_pandas_df(
            self.store.get_artifacts_by_id(artifacts_ids), lambda _artifact: self.get_artifact_df(_artifact)
        )

    def get_one_hop_parent_executions(self, execution_id: t.List[int], pipeline_id: str = None) -> t.List[int]:
        """Get artifacts produced by executions that consume given artifact.

        Args:
            artifact name: Name of an artifact.
        Return:
            Output artifacts of all executions that consumed given artifact.
        """
        artifacts_input=self._get_input_artifacts(execution_id)
        arti = self.store.get_artifacts_by_id(artifacts_input)
        list_exec = []
        exec_ids_added = []
        for i in artifacts_input:
            exec = self._get_executions_by_output_artifact_id(i, pipeline_id)
            if exec not in exec_ids_added:
                exec_ids_added.append(exec)
                list_exec.append(self.store.get_executions_by_id(exec))
        return list_exec

    def get_one_hop_parent_executions_ids(self, execution_ids: t.List[int], pipeline_id: str = None) -> t.List[int]:
        """Get parent execution ids for given execution id
        Args: 
           execution_id : Execution id for which parent execution are required
                          It is passed in list, for example execution_id: [1]  
           pipeline_id : Pipeline id
        Return:
           Returns parent executions for given id
        """
        artifact_ids: t.Optional = self._get_input_artifacts(execution_ids)
        if not artifact_ids:
            return None

        exe_ids = []

        for id in artifact_ids:
            ids = self._get_executions_by_output_artifact_id(id, pipeline_id)
            exe_ids.extend(ids)
        return exe_ids

    def get_executions_with_execution_ids(self, exe_ids: t.List[int]) -> pd.DataFrame:
        """For list of execution ids it returns df with "id,Execution_type_name, Execution_uuid"
        Args:
            execution ids: List of execution ids.
        Return:
            df["id","Execution_type_name","Execution_uuid"]
        """
        df = pd.DataFrame()
        executions = self.store.get_executions_by_id(exe_ids)
        execution_id = {}
        for exe in executions:
            temp_dict = {}
            # To get execution_id, exe list[mlmd.proto.execution] is converted to dict using MessageToDict
            execution_id = MessageToDict(exe, including_default_value_fields=False, preserving_proto_field_name=True)
            temp_dict['id'] = int(execution_id['id'])
            d1 = self._transform_to_dataframe(exe, temp_dict)       # df {id:,executions}
            df = pd.concat([df, d1], sort=True, ignore_index=True)
        df.drop_duplicates()
        df = df[["id", "Execution_type_name","Execution_uuid"]]
        return df

    def get_one_hop_child_executions(self, execution_id: t.List[int]) -> t.List[int]:
        """Get artifacts produced by executions that consume given artifact.

        Args:
            artifact name: Name of an artifact.
        Return:
            Output artifacts of all executions that consumed given artifact.
        """
        artifacts_output=self._get_output_artifacts(execution_id)
        arti=self.store.get_artifacts_by_id(artifacts_output)
        for i in artifacts_output:
            exec=self._get_executions_by_input_artifact_id(i)
            list_exec=self.store.get_executions_by_id(exec)
            for id in list_exec:
                self._transform_to_dataframe(id).Execution_type_name

    def get_all_child_artifacts(self, artifact_name: str) -> pd.DataFrame:
        """Return all downstream artifacts starting from the given artifact.

        Args:
            artifact_name: Artifact name.
        Returns:
            Data frame containing all child artifacts.
        """
        df = pd.DataFrame()
        d1 = self.get_one_hop_child_artifacts(artifact_name)
        # df = df.append(d1, sort=True, ignore_index=True)
        df = pd.concat([df, d1], sort=True, ignore_index=True)
        for row in d1.itertuples():
            d1 = self.get_all_child_artifacts(row.name)
            # df = df.append(d1, sort=True, ignore_index=True)
            df = pd.concat([df, d1], sort=True, ignore_index=True)
        df = df.drop_duplicates(subset=None, keep="first", inplace=False)
        return df

    def get_one_hop_parent_artifacts(self, artifact_name: str) -> pd.DataFrame:
        """Return input artifacts for the execution that produced the given artifact.
        Args:
            artifact_name: Artifact name.
        Returns:
            Data frame containing immediate parent artifactog of given artifact.
        """
        artifact: t.Optional = self._get_artifact(artifact_name)
        if not artifact:
            return pd.DataFrame()

        artifact_ids: t.List[int] = self._get_input_artifacts(self._get_executions_by_output_artifact_id(artifact.id))

        return self._as_pandas_df(
            self.store.get_artifacts_by_id(artifact_ids), lambda _artifact: self.get_artifact_df(_artifact)
        )

    def get_all_parent_artifacts(self, artifact_name: str) -> pd.DataFrame:
        """Return all upstream artifacts.
        Args:
            artifact_name: Artifact name.
        Returns:
            Data frame containing all parent artifacts.
        """
        df = pd.DataFrame()
        d1 = self.get_one_hop_parent_artifacts(artifact_name)
        # df = df.append(d1, sort=True, ignore_index=True)
        df = pd.concat([df, d1], sort=True, ignore_index=True)
        for row in d1.itertuples():
            d1 = self.get_all_parent_artifacts(row.name)
            # df = df.append(d1, sort=True, ignore_index=True)
            df = pd.concat([df, d1], sort=True, ignore_index=True)
        df = df.drop_duplicates(subset=None, keep="first", inplace=False)
        return df

    def get_all_parent_executions_by_id(self, execution_id: t.List[int], pipeline_id: str = None) -> t.List[int]:
        parent_executions = [[],[]]
        current_execution_id = execution_id
        list_of_parent_execution_id = []
        link_src_trgt_list = []
        while current_execution_id:
            parent_execution_ids = self.get_one_hop_parent_executions(current_execution_id, pipeline_id)
            list_of_parent_execution_id = []
            for data in parent_execution_ids:
                for j in data:
                    temp=[j.id, j.properties["Execution_type_name"].string_value, j.properties["Execution_uuid"].string_value]
                    if temp not in parent_executions[0]:
                        link_src_trgt_list.append({"source":j.id, "target":current_execution_id[0]})
                        list_of_parent_execution_id.append(temp)
            if list_of_parent_execution_id:
                parent_executions[0].extend(list_of_parent_execution_id)
                parent_executions[1].extend(link_src_trgt_list)
                for id_name_uuid in list_of_parent_execution_id:
                    current_execution_id = [id_name_uuid[0]]
                    recursive_parents = self.get_all_parent_executions_by_id(current_execution_id, pipeline_id)
                    parent_executions[0].extend(recursive_parents[0])
                    parent_executions[1].extend(recursive_parents[1])
            else:
                break
        return parent_executions

    def get_all_parent_executions(self, artifact_name: str) -> pd.DataFrame:
        """Return all executions that produced upstream artifacts for the given artifact.
        Args:
            artifact_name: Artifact name.
        Returns:
            Data frame containing all parent executions.
        """
        parent_artifacts: pd.DataFrame = self.get_all_parent_artifacts(artifact_name)
        if parent_artifacts.shape[0] == 0:
            # If it's empty, there's no `id` column and the code below raises an exception.
            return pd.DataFrame()

        execution_ids = set(
            event.execution_id
            for event in self.store.get_events_by_artifact_ids(parent_artifacts.id.values.tolist())
            if event.type == mlpb.Event.OUTPUT
        )

        return self._as_pandas_df(
            self.store.get_executions_by_id(execution_ids),
            lambda _exec: self._transform_to_dataframe(_exec, {"id": _exec.id, "name": _exec.name}),
        )

    def find_producer_execution(self, artifact_name: str) -> t.Optional[mlpb.Execution]:
        """Return execution that produced the given artifact.

        One artifact can have multiple producer executions (names of artifacts are fully-qualified with hashes). So,
        if two executions produced the same exact artifact, this one artifact will have multiple parent executions.
        """
        artifact: t.Optional[mlpb.Artifact] = self._get_artifact(artifact_name)
        if not artifact:
            logger.debug("Artifact does not exist (name=%s).", artifact_name)
            return None

        executions_ids = set(
            event.execution_id
            for event in self.store.get_events_by_artifact_ids([artifact.id])

            if event.type == mlpb.Event.OUTPUT
        )
        if not executions_ids:
            logger.debug("No producer execution exists for artifact (name=%s, id=%s).", artifact.name, artifact.id)
            return None

        executions: t.List[mlpb.Execution] = self.store.get_executions_by_id(executions_ids)
        if not executions:
            logger.debug("No executions exist for given IDs (ids=%s)", str(executions_ids))
            return None

        if len(executions) >= 2:
            logger.debug(
                "Multiple executions (ids=%s) claim artifact (name=%s) as output.",
                [e.id for e in executions],
                artifact.name,
            )

        return executions[0]

    get_producer_execution = find_producer_execution

    def get_metrics(self, metrics_name: str) -> t.Optional[pd.DataFrame]:
        """Return metric data frame.
        Args:
            metrics_name: Metrics name.
        Returns:
            Data frame containing all metrics.
        """
        for metric in self.store.get_artifacts_by_type("Step_Metrics"):
            if metric.name == metrics_name:
                name: t.Optional[str] = metric.custom_properties.get("Name", None)
                if name:
                    return pd.read_parquet(name)
                break
        return None

    @staticmethod
    def read_dataslice(name: str) -> pd.DataFrame:
        """Reads the data slice."""
        # To do checkout if not there
        df = pd.read_parquet(name)
        return df


    # writing new functions to remove multiple calls to cmfquery functions or ml-metadata functions
    def get_all_executions_in_pipeline(self, pipeline_name: str) -> pd.DataFrame:
        """Return all executions of the given pipeline as pandas data frame.
        Args:
            pipeline_name:- pipeline name
        Returns:
            Data frame with all executions associated with the given pipeline.
        """
        df = pd.DataFrame()
        pipeline_id = self.get_pipeline_id(pipeline_name)
        for stage in self._get_stages(pipeline_id):
            for execution in self._get_executions(stage.id):
               ex_as_df: pd.DataFrame = self._transform_to_dataframe(
                   execution, {"id": execution.id, "name": execution.name}
               )
               df = pd.concat([df, ex_as_df], sort=True, ignore_index=True)
        return df

    def get_all_artifacts_for_executions(self, execution_ids: t.List[int]) -> pd.DataFrame:
        """Return all artifacts for the list of given executions.

        Args:
            execution_ids: List of Execution identifiers.
        Return:
            Data frame containing artifacts for the list of given executions.
        """
        df = pd.DataFrame()
        # set of artifact ids for list of given execution ids
        artifact_ids = set(
            event.artifact_id
            for event in self.store.get_events_by_execution_ids(set(execution_ids))
            )
        artifacts = self.store.get_artifacts_by_id(list(artifact_ids))
        for artifact in artifacts:
             df = pd.concat(
                    [df, self.get_artifact_df(artifact)], sort=True, ignore_index=True
             )
        return df
    
    def get_one_hop_parent_artifacts_with_id(self, artifact_id: int) -> pd.DataFrame:
        """Return input artifacts for the execution that produced the given artifact.
        Args:
            artifact_id: Artifact id.
        Returns:
            Data frame containing immediate parent artifacts of given artifact/artifacts.
        """
        df = pd.DataFrame()
        input_artifact_ids: t.List[int] = self._get_input_artifacts(self._get_executions_by_output_artifact_id(artifact_id))
        df = self._as_pandas_df(self.store.get_artifacts_by_id(input_artifact_ids), 
                lambda _artifact: self.get_artifact_df(_artifact)
                )
        return df

    def dumptojson(self, pipeline_name: str, exec_id: t.Optional[int] = None) -> t.Optional[str]:
        """Return JSON-parsable string containing details about the given pipeline.
        Args:
            pipeline_name: Name of an AI pipelines.
            exec_id: Optional stage execution ID - filter stages by this execution ID.
        Returns:
            Pipeline in JSON format.
        """
        if exec_id is not None:
            exec_id = int(exec_id)

        def _get_node_attributes(_node: t.Union[mlpb.Context, mlpb.Execution, mlpb.Event], _attrs: t.Dict) -> t.Dict:
            for attr in CONTEXT_LIST:
                #Artifacts getattr call on Type was giving empty string, which was overwriting 
                # the defined types such as Dataset, Metrics, Models
                if getattr(_node, attr, None) is not None and not getattr(_node, attr, None) == "":
                    _attrs[attr] = getattr(_node, attr)

            if "properties" in _attrs:
                _attrs["properties"] = CmfQuery._copy(_attrs["properties"])
            if "custom_properties" in _attrs:
                # TODO: (sergey) why do we need to rename "type" to "user_type" if we just copy into a new dictionary?
                _attrs["custom_properties"] = CmfQuery._copy(
                    _attrs["custom_properties"], key_mapper={"type": "user_type"}
                )
            return _attrs

        pipelines: t.List[t.Dict] = []
        for pipeline in self._get_pipelines(pipeline_name):
            pipeline_attrs = _get_node_attributes(pipeline, {"stages": []})
            for stage in self._get_stages(pipeline.id):
                stage_attrs = _get_node_attributes(stage, {"executions": []})
                for execution in self._get_executions(stage.id, execution_id=exec_id):
                    # name will be an empty string for executions that are created with
                    # create new execution as true(default)
                    # In other words name property will there only for execution
                    # that are created with create new execution flag set to false(special case)
                    exec_attrs = _get_node_attributes(
                        execution,
                        {
                            "type": self.store.get_execution_types_by_id([execution.type_id])[0].name,
                            "name": execution.name if execution.name != "" else "",
                            "events": [],
                        },
                    )
                    for event in self.store.get_events_by_execution_ids([execution.id]):
                        event_attrs = _get_node_attributes(event, {})
                        # An event has only a single Artifact associated with it. 
                        # For every artifact we create an event to link it to the execution.

                        artifacts =  self.store.get_artifacts_by_id([event.artifact_id])
                        artifact_attrs = _get_node_attributes(
                                artifacts[0], {"type": self.store.get_artifact_types_by_id([artifacts[0].type_id])[0].name}
                            )
                        event_attrs["artifact"] = artifact_attrs
                        exec_attrs["events"].append(event_attrs)
                    stage_attrs["executions"].append(exec_attrs)
                pipeline_attrs["stages"].append(stage_attrs)
            pipelines.append(pipeline_attrs)

        return json.dumps({"Pipeline": pipelines})

    """def materialize(self, artifact_name:str):
       artifacts = self.store.get_artifacts()
       for art in artifacts:
           if art.name == artifact_name:
               selected_artifact = art
               break
       for k, v in selected_artifact.custom_properties.items():
           if (k == "Path"):
               path = v
           elif (k == "git_repo"):
               git_repo = v
           elif (k == "Revision"):
               rev = v
           elif (remote == "Remote"):
               remote = v
       
       Cmf.materialize(path, git_repo, rev, remote)"""


def test_on_collision() -> None:
    from unittest import TestCase

    tc = TestCase()

    tc.assertEqual(3, len(_KeyMapper.OnCollision))
    tc.assertEqual(0, _KeyMapper.OnCollision.DO_NOTHING.value)
    tc.assertEqual(1, _KeyMapper.OnCollision.RESOLVE.value)
    tc.assertEqual(2, _KeyMapper.OnCollision.RAISE_ERROR.value)


def test_dict_mapper() -> None:
    from unittest import TestCase

    tc = TestCase()

    dm = _DictMapper({"src_key": "tgt_key"}, on_collision=_KeyMapper.OnCollision.RESOLVE)
    tc.assertEqual("tgt_key", dm.get({}, "src_key"))
    tc.assertEqual("other_key", dm.get({}, "other_key"))
    tc.assertEqual("existing_key_1", dm.get({"existing_key": "value"}, "existing_key"))
    tc.assertEqual("existing_key_2", dm.get({"existing_key": "value", "existing_key_1": "value_1"}, "existing_key"))

    dm = _DictMapper({"src_key": "tgt_key"}, on_collision=_KeyMapper.OnCollision.DO_NOTHING)
    tc.assertEqual("existing_key", dm.get({"existing_key": "value"}, "existing_key"))


def test_prefix_mapper() -> None:
    from unittest import TestCase

    tc = TestCase()

    pm = _PrefixMapper("nested_", on_collision=_KeyMapper.OnCollision.RESOLVE)
    tc.assertEqual("nested_src_key", pm.get({}, "src_key"))

    tc.assertEqual("nested_existing_key_1", pm.get({"nested_existing_key": "value"}, "existing_key"))
    tc.assertEqual(
        "nested_existing_key_2",
        pm.get({"nested_existing_key": "value", "nested_existing_key_1": "value_1"}, "existing_key"),
    )

    dm = _PrefixMapper("nested_", on_collision=_KeyMapper.OnCollision.DO_NOTHING)
    tc.assertEqual("nested_existing_key", dm.get({"nested_existing_key": "value"}, "existing_key"))


if __name__ == "__main__":
    test_on_collision()
    test_dict_mapper()
    test_prefix_mapper()



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import subprocess
import os
import sys

def fix_subparsers(subparsers):
    subparsers.required = True
    subparsers.dest = "cmd"

def git_exists():
    try:
        output = subprocess.check_output(["git", "version"]).decode("ascii")
        return output
    except FileNotFoundError:
        return "ERROR: git is not installed!! Install git."


def find_root(file_name: str):
    """This function returns the root of the file passed as a parameter.
    It searches for the file in the parent directory until the mount
    if it is not in the current directory.
    """
    msg = "'cmf' is not configured.\nExecute 'cmf init' command."
    # make current directory as root directory
    root_dir = os.path.realpath(os.getcwd())
    while True:
        # adding file name to the root directory
        file_path = os.path.join(root_dir, file_name)
        # whether file path exists, if yes return function with root_dir
        if os.path.exists(file_path):
            return root_dir
        # as file doesn't exists in root_dir and if root_dir is mount directory return msg
        if os.path.ismount(root_dir):
            return msg
        # assigning parent of root_dir
        parent = os.path.abspath(os.path.join(root_dir, os.pardir))
        # updating root_dir with its parent
        root_dir = parent


def check_minio_server(dvc_config_op):
    from minio import Minio
    from minio.error import S3Error

    if dvc_config_op["core.remote"] == "minio":
        endpoint = dvc_config_op["remote.minio.endpointurl"].split("http://")[1]
        access_key = dvc_config_op["remote.minio.access_key_id"]
        secret_key = dvc_config_op["remote.minio.secret_access_key"]
        bucket_name = dvc_config_op["remote.minio.url"].split("s3://")[1]
        try:
            client = Minio(
                endpoint, access_key=access_key, secret_key=secret_key, secure=False
            )
            found = client.bucket_exists(bucket_name)
            if found:
                return "SUCCESS"
        except TypeError as exception:
            return exception
        except S3Error as exception:
            return exception



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import requests
#import urllib3
#urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class OSDFremoteArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        host: str, #s_url 
        current_directory: str, #current_directory where cmf artifact pull is executed
        remote_file_path: str, # download_loc of the artifact 
        local_path: str, #name of the artifact
    ):
        output = ""
        remote_repo = dvc_config_op["remote.osdf.url"]
        user = "nobody"
        dynamic_password = dvc_config_op["remote.osdf.password"]
        custom_auth_header = dvc_config_op["remote.osdf.custom_auth_header"]
        #print(f"dynamic password from download_artifacts={dynamic_password}")
        #print(f"Fetching artifact={local_path}, surl={host} to {remote_file_path} when this has been called at {current_directory}")

        try:
            headers={dvc_config_op["remote.osdf.custom_auth_header"]: dvc_config_op["remote.osdf.password"]}
            temp = local_path.split("/")
            temp.pop()
            dir_path = "/".join(temp)
            dir_to_create = os.path.join(current_directory, dir_path)
            os.makedirs(
                dir_to_create, mode=0o777, exist_ok=True
            )  # creates subfolders needed as per artifacts folder structure
            local_file_path = os.path.join(current_directory, local_path)
            local_file_path = os.path.abspath(local_file_path)

            response = requests.get(host, headers=headers, verify=True) #This should be made True. otherwise this will produce Insecure SSL Warning
            if response.status_code == 200 and response.content:
                data = response.content
            else:
                return "No data received from the server."

        except Exception as exception:
            return exception

        try:
            with open(remote_file_path, 'wb') as file:
                file.write(data)
            if os.path.exists(remote_file_path) and os.path.getsize(remote_file_path) > 0:
                #print(f"object {local_path} downloaded at {remote_file_path}")
                stmt = f"object {local_path} downloaded at {remote_file_path}."
                return stmt
        except Exception as e:
            print(f"An error occurred while writing to the file: {e}")



###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import subprocess
import dvc.api
import dvc.exceptions
from typing import List, Optional

def check_git_remote() -> bool:
    process = ""
    commit = ""
    git_remote_configured = False
    try:
        process = subprocess.Popen(['git', 'remote', 'show'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # output = process.stdout.readline()
        output, error = process.communicate(timeout=60)

        remote = output.strip()
        if remote:
            git_remote_configured = True
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
    return git_remote_configured


def check_default_remote() -> bool:
    process = ""
    commit = ""
    dvc_configured = False
    try:
        process = subprocess.Popen(['dvc', 'config', 'core.remote'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # output = process.stdout.readline()
        output, error = process.communicate(timeout=60)

        remote = output.strip()
        if remote:
            dvc_configured = True
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
    return dvc_configured


def dvc_get_url(folder: str, retry: bool = False, repo: str = "") -> str:
    url = ""
    try:
        if not repo and not repo.isspace():
            url = dvc.api.get_url(folder)
        else:
            url = dvc.api.get_url(folder, repo)
    except dvc.exceptions.PathMissingError as err:
        if not retry:
            print(f"Retrying with full path")
            folder = os.path.join(os.getcwd(), folder)
            url = dvc_get_url(folder, True)
        else:
            print(f"dvc.exceptions.PathMissingError Caught  Unexpected {err}, {type(err)}")
    except dvc.exceptions.OutputNotFoundError as err:
        if not retry:
            filename = folder.split('/')[-1]
            folder = os.path.join(os.getcwd() , filename)
            url = dvc_get_url(folder, True)

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
    return url


def dvc_get_hash(folder: str, repo: str = "") -> str:
    c_hash = ""
    try:
        url = dvc_get_url(folder, False, repo)
        url_list = url.split('/')
        len_list = len(url_list)
        c_hash = ''.join(url_list[len_list - 2:len_list])

    except dvc.exceptions.PathMissingError as err:
        print(f"dvc.exceptions.PathMissingError Caught  Unexpected {err}, {type(err)}")
    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
    return c_hash


def check_git_repo() -> bool:

    process = ""
    commit = ""
    is_git_repo = False
    try:
        process = subprocess.Popen(['git',
                                    'rev-parse',
                                    '--is-inside-work-tree'],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   universal_newlines=True)
        # output = process.stdout.readline()
        output, error = process.communicate(timeout=60)

        is_git_repo = output.strip()
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
    return is_git_repo


def git_checkout_new_branch(branch_name: str):

    process = ""
    commit = ""
    try:
        process = subprocess.Popen(['git',
                                    'checkout',
                                    '-q',
                                    '-B',
                                    branch_name],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # output = process.stdout.readline()
        output, error = process.communicate(timeout=60)

        commit = output.strip()
        print(f"*** Note: CMF will check out a new branch in git to commit the metadata files ***\n"
              f"*** The checked out branch is {branch_name}. ***")
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
        
        print(f"Unexpected {err}, {type(err)}")
        print(f"Unexpected {outs}")
        print(f"Unexpected {errs}")
        print(f"Checking out new branch for the execution failed, continuing in the default branch.")


def git_get_commit() -> str:
    process = ""
    commit = ""
    try:
        process = subprocess.Popen(['git', 'rev-parse', 'HEAD'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # output = process.stdout.readline()
        output, error = process.communicate(timeout=60)
        commit = output.strip()
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
        print(f"Unexpected {err}, {type(err)}")
        print(f"Unexpected {outs}")
        print(f"Unexpected {errs}")
    return commit


def commit_dvc_lock_file(file_path: str, execution_id) -> str:
    commit = ""
    process = ""
    try:
        process = subprocess.Popen(['git', 'add', file_path],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # To-Do : Parse the output and report if error
        _, _ = process.communicate(timeout=60)
        process = subprocess.Popen(
            [
                'git',
                'commit',
                '-m ' +
                'commiting ' +
                str(file_path) +
                "-" +
                str(execution_id)],
            stdout=subprocess.PIPE,
            universal_newlines=True)

        output, errs = process.communicate(timeout=60)
        commit = output.strip()
        process = subprocess.Popen(['git', 'log', file_path],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # To-Do : Parse the output and report if error
        output, errs = process.communicate(timeout=60)
        commit = output.splitlines()[0].strip()
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
        print(f"Unexpected {err}, {type(err)}")
        print(f"Unexpected {outs}")
        print(f"Unexpected {errs}")
    return commit

def git_commit(execution_id: str) -> str:
    commit = ""
    process = None
    try:
        # To-Do : Parse the output and report if error
        process = subprocess.Popen(['git', 'commit', '-m ' + 'commiting ' + str(execution_id)],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)

        output, errs = process.communicate(timeout=60)
        commit = output.strip()

        process = subprocess.Popen(['git', 'log'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        # To-Do : Parse the output and report if error
        output, errs = process.communicate(timeout=60)
        commit = output.splitlines()[0].strip()
    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit

def commit_output(folder: str, execution_id: str) -> str:
    commit = ""
    process = ""
    try:
        if os.path.exists(os.getcwd() + '/' + folder):
            sub_dir_file = True
            process = subprocess.Popen(['dvc', 'add', folder],
                                    stdout=subprocess.PIPE,
                                    universal_newlines=True)
        else:
            sub_dir_file = False
            process = subprocess.Popen(['dvc', 'import-url', '--to-remote', folder],
                                stdout=subprocess.PIPE,
                                universal_newlines=True)
            
        output, errs = process.communicate()
        
        if process.returncode != 0:
            raise Exception(f'DVC add/import-url failed, Check if DVC is tracking parent directory: {errs}')
        
        commit = output.strip()

        if sub_dir_file:
            process = subprocess.Popen(['git', 'add', folder + '.dvc'],
                                    stdout=subprocess.PIPE,
                                    universal_newlines=True)
        else:
            process = subprocess.Popen(['git', 'add', folder.split('/')[-1] + '.dvc'],
                                        stdout=subprocess.PIPE,
                                        universal_newlines=True)
        
        
        output, errs = process.communicate(timeout=60)
        
        if process.returncode != 0:
            raise Exception(f"Git add failed, Check gitignore: {errs}")
        
    except Exception as err:
        process.kill()
        outs, errs = process.communicate()
        print(f"Unexpected {err}, {type(err)}")
        print(f"Unexpected {outs}")
        print(f"Unexpected {errs}")
    return commit


# Get the remote repo
def git_get_repo() -> str:
    commit = ""
    process = ""
    output = ""
    errs = ""
    try:
        process = subprocess.Popen(['git', 'remote', '-v'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        process.kill()
        print(f"Unexpected {err}, {type(err)}")
        print(f"Unexpected {output}")
        print(f"Unexpected {errs}")
    return commit.split()[1]

#Initialise git with quiet option
def git_quiet_init() -> str:
    commit = ""
    try:
        process = subprocess.Popen(['git', 'init', '-q'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit


# Initial git commit
def git_initial_commit() -> str:
    commit = ""
    try:
        process = subprocess.Popen(['git', 'commit', '-q', '--allow-empty', '-n', '-m', "Initial code commit"],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit

# Add a remote repo url
def git_add_remote(git_url) -> str:
    commit = ""
    try:
        process = subprocess.Popen(['git', 'remote', 'add', 'cmf_origin', f"{git_url}"],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit

# dvc init with quiet option
def dvc_quiet_init() -> str:
    commit = ""
    try:
        process = subprocess.Popen(['dvc', 'init', '-q'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit

# add repo in dvc
def dvc_add_remote_repo(repo_type, repo_path) -> str:
    commit = ""
    try:
        process = subprocess.Popen(['dvc', 'remote', 'add', '-d', '-f', f"{repo_type}", f"{repo_path}"],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit

# add repo related attributes in dvc
def dvc_add_attribute(repo_type, attribute_type, attribute_value) -> str:
    commit = ""
    try:
        process = subprocess.Popen(['dvc', 'remote', 'modify', f"{repo_type}", f"{attribute_type}", f"{attribute_value}"],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit


# get dvc config
def dvc_get_config() -> str:
    commit = ""
    try:
        process = subprocess.Popen(['dvc','config', '-l'],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
        output, errs = process.communicate(timeout=60)
        commit = output.strip()

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")
        if isinstance(object, subprocess.Popen):
           process.kill()
           outs, errs = process.communicate()
           print(f"Unexpected {outs}")
           print(f"Unexpected {errs}")
    return commit


# dvc push
def dvc_push(file_list: Optional[List[str]] = None) -> str:
    commit = ""
    if file_list is None:
       try:
           process = subprocess.Popen(['dvc', 'push'],
                                  stdout=subprocess.PIPE,
                                   universal_newlines=True)
           output, errs = process.communicate()
           commit = output.strip()

       except Exception as err:
           print(f"Unexpected {err}, {type(err)}")
           if isinstance(object, subprocess.Popen):
              process.kill()
              outs, errs = process.communicate()
              print(f"Unexpected {outs}")
              print(f"Unexpected {errs}")

    else:
        file_list.insert(0, 'dvc')
        file_list.insert(1, 'push')
        try:
            process = subprocess.Popen(file_list,
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)
            output, errs = process.communicate()
            commit = output.strip()

        except Exception as err:
           print(f"Unexpected {err}, {type(err)}")
           if isinstance(object, subprocess.Popen):
              process.kill()
              outs, errs = process.communicate()
              print(f"Unexpected {outs}")
              print(f"Unexpected {errs}")
    return commit



# cmf-server api's
from fastapi import FastAPI, Request, status, HTTPException, Query, UploadFile, File
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse, StreamingResponse
from fastapi.staticfiles import StaticFiles
from contextlib import asynccontextmanager
import pandas as pd
from typing import List, Dict, Any

from cmflib import cmfquery, cmf_merger
from server.app.get_data import (
    get_artifacts,
    get_lineage_data,
    create_unique_executions,
    get_mlmd_from_server,
    get_artifact_types,
    get_all_artifact_ids,
    get_all_exe_ids,
    get_executions,
    get_model_data
)
from server.app.query_artifact_lineage_d3force import query_artifact_lineage_d3force
from server.app.query_execution_lineage_d3force import query_execution_lineage_d3force
from server.app.query_execution_lineage_d3tree import query_execution_lineage_d3tree
from server.app.query_artifact_lineage_d3tree import query_artifact_lineage_d3tree
from pathlib import Path
import os
import json

server_store_path = "/cmf-server/data/mlmd"

dict_of_art_ids = {}
dict_of_exe_ids = {}

#lifespan used to prevent multiple loading and save time for visualization.
@asynccontextmanager
async def lifespan(app: FastAPI):
    global dict_of_art_ids
    global dict_of_exe_ids
    if os.path.exists(server_store_path):
        # loaded execution ids with names into memory
        dict_of_exe_ids = await get_all_exe_ids(server_store_path)
        # loaded artifact ids into memory
        dict_of_art_ids = await get_all_artifact_ids(server_store_path, dict_of_exe_ids)
    yield
    dict_of_art_ids.clear()
    dict_of_exe_ids.clear()

app = FastAPI(title="cmf-server", lifespan=lifespan)

BASE_PATH = Path(__file__).resolve().parent
app.mount("/cmf-server/data/static", StaticFiles(directory="/cmf-server/data/static"), name="static")
server_store_path = "/cmf-server/data/mlmd"

my_ip = os.environ.get("MYIP", "127.0.0.1")
hostname = os.environ.get('HOSTNAME', "localhost")

#checking if IP or Hostname is provided,initializing url accordingly.
if my_ip != "127.0.0.1":
    url="http://"+my_ip+":3000"
else:
    url="http://"+hostname+":3000"

origins = [
    "http://localhost:3000",
    "localhost:3000",
    url
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/")
async def read_root(request: Request):
    return {"cmf-server"}

# api to post mlmd file to cmf-server
@app.post("/mlmd_push")
async def mlmd_push(info: Request):
    print("mlmd push started")
    print("......................")
    req_info = await info.json()
    pipeline_name = req_info["pipeline_name"]
    status = await create_unique_executions(server_store_path, req_info)
    if status == "version_update":
        # Raise an HTTPException with status code 422
        raise HTTPException(status_code=422, detail="version_update")
    # async function
    await update_global_exe_dict(pipeline_name)
    await update_global_art_dict(pipeline_name)
    return {"status": status, "data": req_info}


# api to get mlmd file from cmf-server
@app.get("/mlmd_pull/{pipeline_name}", response_class=HTMLResponse)
async def mlmd_pull(info: Request, pipeline_name: str):
    # checks if mlmd file exists on server
    req_info = await info.json()
    if os.path.exists(server_store_path):
        #json_payload values can be json data, NULL or no_exec_id.
        json_payload= await get_mlmd_from_server(server_store_path, pipeline_name, req_info['exec_id'])
    else:
        print("No mlmd file submitted.")
        json_payload = ""
    return json_payload

# api to display executions available in mlmd
@app.get("/executions/{pipeline_name}")
async def executions(
    request: Request,
    pipeline_name: str,
    page: int = Query(1, description="Page number", gt=0),
    per_page: int = Query(5, description="Items per page", le=100),
    sort_field: str = Query("Context_Type", description="Column to sort by"),
    sort_order: str = Query("asc", description="Sort order (asc or desc)"),
    filter_by: str = Query(None, description="Filter by column"),
    filter_value: str = Query(None, description="Filter value"),
    ):
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        exe_ids_initial = dict_of_exe_ids[pipeline_name]
        # Apply filtering if provided
        if filter_by and filter_value:
            exe_ids_initial = exe_ids_initial[exe_ids_initial[filter_by].str.contains(filter_value, case=False)]
        # Apply sorting if provided
        exe_ids_sorted = exe_ids_initial.sort_values(by=sort_field, ascending=(sort_order == "asc"))
        exe_ids = exe_ids_sorted['id'].tolist()
        total_items = len(exe_ids)
        start_idx = (page - 1) * per_page
        end_idx = start_idx + per_page
        if total_items < end_idx:
            end_idx = total_items
        exe_ids_list = exe_ids[start_idx:end_idx]
        executions_df = await get_executions(server_store_path, pipeline_name, exe_ids_list)
        temp = executions_df.to_json(orient="records")
        executions_parsed = json.loads(temp)
        return {
            "total_items": total_items,
            "items": executions_parsed
        }
    else:
        return

@app.get("/artifact-lineage/force-directed-graph/{pipeline_name}")
async def artifact_lineage(request: Request, pipeline_name: str):
    '''
      This api returns dictionary of nodes and links for given pipeline.
      response = {
                   nodes: [{id:"",name:""}],
                   links: [{source:1,target:4},{}],
                 }

    '''
    # checks if mlmd file exists on server
 
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response=await get_lineage_data(server_store_path,pipeline_name,"Artifacts",dict_of_art_ids,dict_of_exe_ids)
            #response = None
            return response
        else:
            return f"Pipeline name {pipeline_name} doesn't exist."

    else:
        return None

@app.get("/list-of-executions/{pipeline_name}")
async def list_of_executions(request: Request, pipeline_name: str):
    '''
      This api's returns list of execution types.

    '''
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response = await get_lineage_data(server_store_path,pipeline_name,"Execution",dict_of_art_ids,dict_of_exe_ids)
            return response
        else:
            return f"Pipeline name {pipeline_name} doesn't exist."

    else:
        return None

@app.get("/execution-lineage/force-directed-graph/{pipeline_name}/{uuid}")
async def execution_lineage(request: Request, pipeline_name: str, uuid: str):
    '''
      returns dictionary of nodes and links for given execution_type.
      response = {
                   nodes: [{id:"",name:"",execution_uuid:""}],
                   links: [{source:1,target:4},{}],
                 } 
    '''
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response = await query_execution_lineage_d3force(server_store_path, pipeline_name, dict_of_exe_ids, uuid)
    else:
        response = None
    return response
    
@app.get("/execution-lineage/tangled-tree/{uuid}/{pipeline_name}")
async def execution_lineage(request: Request,uuid, pipeline_name: str):
    '''
      returns dictionary of nodes and links for given execution_type.
      response = {
                   nodes: [{id:"",name:"",execution_uuid:""}],
                   links: [{source:1,target:4},{}],
                 } 
    '''
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response = await query_execution_lineage_d3tree(server_store_path, pipeline_name, dict_of_exe_ids,uuid)
    return response

# api to display artifacts available in mlmd
@app.get("/artifacts/{pipeline_name}/{type}")
async def artifacts(
    request: Request,
    pipeline_name: str,
    type: str,   # type = artifact type
    page: int = Query(1, description="Page number", gt=0),
    per_page: int = Query(5, description="Items per page", le=100),
    sort_field: str = Query("name", description="Column to sort by"),
    sort_order: str = Query("asc", description="Sort order (asc or desc)"),
    filter_by: str = Query(None, description="Filter by column"),
    filter_value: str = Query(None, description="Filter value"),
    ):
    empty_df = pd.DataFrame()
    art_ids_dict = {}
    art_type = type
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        art_ids_dict = dict_of_art_ids[pipeline_name]
        if not art_ids_dict:
            return {               #return {items: None} so that GUI loads 
            "total_items": 0,      #empty page when art_ids_dict is {}
            "items": None
            }
        art_ids_initial = []
        if art_type in art_ids_dict:
            art_ids_initial = art_ids_dict[art_type]
        else:
            return {
            "total_items": 0,
            "items": None
        }
        # Apply filtering if provided
        if filter_by and filter_value:
            art_ids_initial = art_ids_initial[art_ids_initial[filter_by].str.contains(filter_value, case=False)]
        # Apply sorting if provided
        art_ids_sorted = art_ids_initial.sort_values(by=sort_field, ascending=(sort_order == "asc"))
        art_ids = art_ids_sorted['id'].tolist()
        total_items = len(art_ids)
        start_idx = (page - 1) * per_page
        end_idx = start_idx + per_page
        if total_items < end_idx:
            end_idx = total_items
        artifact_id_list = list(art_ids)[start_idx:end_idx]
        artifact_df = await get_artifacts(server_store_path, pipeline_name, art_type, artifact_id_list)
        data_paginated = artifact_df
        #data_paginated is returned None if artifact df is None or {}
        #it will load empty page, without this condition it will load 
        #data of whichever artifact_type is loaded before this. 
        if artifact_df == None or artifact_df == {}:   
            data_paginated = None      
            total_items = 0
        return {
            "total_items": total_items,
            "items": data_paginated
        }
    else:
        print(f"{server_store_path} file doesn't exist.")
        return {
            "total_items": 0,
            "items": None
        }

@app.get("/artifact-lineage/tangled-tree/{pipeline_name}")
async def artifact_lineage(request: Request, pipeline_name: str) -> List[List[Dict[str, Any]]]:
    '''
      Returns:
      A nested list of dictionaries with 'id' and 'parents' keys.
      response = [
        [{'id': 'data.xml.gz:236d', 'parents': []}],
        [{'id': 'parsed/train.tsv:32b7', 'parents': ['data.xml.gz:236d']}, 
        ]
    '''
    # checks if mlmd file exists on server
    response = None
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response = await query_artifact_lineage_d3tree(server_store_path, pipeline_name, dict_of_art_ids)
            #response = "null"
    return response

#This api's returns list of artifact types.
@app.get("/artifact_types")
async def artifact_types(request: Request):
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        artifact_types = get_artifact_types(server_store_path)
        return artifact_types
    else:
        artifact_types = ""
        return


@app.get("/pipelines")
async def pipelines(request: Request):
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        pipeline_names = query.get_pipeline_names()
        return pipeline_names
    else:
        print("No mlmd file submitted.")
        pipeline_names = []
        return pipeline_names


@app.post("/tensorboard")
async def upload_file(request:Request, pipeline_name: str = Query(..., description="Pipeline name"),
    file: UploadFile = File(..., description="The file to upload")):
    try:
        file_path = os.path.join("/cmf-server/data/tensorboard-logs", pipeline_name, file.filename)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as buffer:
            buffer.write(await file.read())
        return {"message": f"File '{file.filename}' uploaded successfully"}
    except Exception as e:
        return {"error": f"Failed to up load file: {e}"}

@app.get("/model-card")
async def model_card(request:Request, modelId: int, response_model=List[Dict[str, Any]]):
    json_payload_1 = ""
    json_payload_2 = ""
    json_payload_3 = ""
    json_payload_4 = ""
    model_data_df = pd.DataFrame()
    model_exe_df = pd.DataFrame()
    model_input_art_df = pd.DataFrame()
    model_output_art_df = pd.DataFrame()
    df = pd.DataFrame()
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        model_data_df, model_exe_df, model_input_art_df, model_output_art_df  = await get_model_data(server_store_path, modelId)
        if not model_data_df.empty:
            result_1 = model_data_df.to_json(orient="records")
            json_payload_1 = json.loads(result_1)
        if not model_exe_df.empty:
            result_2 = model_exe_df.to_json(orient="records")
            json_payload_2 = json.loads(result_2)
        if not model_input_art_df.empty:
            result_3 =  model_input_art_df.to_json(orient="records")
            json_payload_3 = json.loads(result_3)
        if not model_output_art_df.empty:
            result_4 =  model_output_art_df.to_json(orient="records")
            json_payload_4 = json.loads(result_4)
    return [json_payload_1, json_payload_2, json_payload_3, json_payload_4]

async def update_global_art_dict(pipeline_name):
    global dict_of_art_ids
    output_dict = await get_all_artifact_ids(server_store_path, dict_of_exe_ids, pipeline_name)
    # type(dict_of_exe_ids[pipeline_name]) = Dict[ <class 'pandas.core.frame.DataFrame'> ]
    dict_of_art_ids[pipeline_name]=output_dict[pipeline_name]
    return


async def update_global_exe_dict(pipeline_name):
    global dict_of_exe_ids
    output_dict = await get_all_exe_ids(server_store_path, pipeline_name)
    # type(dict_of_exe_ids[pipeline_name]) = <class 'pandas.core.frame.DataFrame'>
    dict_of_exe_ids[pipeline_name] = output_dict[pipeline_name]  
    return



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import sys
import subprocess
import json

def is_url(url)-> bool:
    from urllib.parse import urlparse
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False

def is_git_repo():
    git_dir = os.path.join(os.getcwd(), '.git')
    print("git_dir", git_dir)
    result = os.path.exists(git_dir) and os.path.isdir(git_dir)
    if result:
        return f"A Git repository already exists in {git_dir}."
    else:
        return


def get_python_env()-> str:
    installed_packages = ""
    python_version = sys.version
    packages = ""
    # check if conda is installed
    if is_conda_installed():
        import conda
        # List all installed packages and their versions
        data = list_conda_packages_json()
        transformed_result = [f"{entry['name']}=={entry['version']}" for entry in data]
        installed_packages =  transformed_result
        packages = f"Conda: Python {python_version}: {installed_packages}"
    else:
        # pip
        try:
            from pip._internal.operations import freeze

            # List all installed packages and their versions
            installed_packages_generator = freeze.freeze()
            installed_packages = list(installed_packages_generator)
            packages = f"Python {python_version}: {installed_packages}"
        except ImportError:
            print("Pip is not installed.")
    return packages

def change_dir(cmf_init_path):
    logging_dir = os.getcwd()
    if not logging_dir == cmf_init_path:
        os.chdir(cmf_init_path)
    return logging_dir

def is_conda_installed():
    try:
        import conda
        # Run the 'conda --version' command and capture the output
        subprocess.run(['conda', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except subprocess.CalledProcessError:
        return False
    except ImportError:
        return False


def list_conda_packages_json():
    try:
        result = subprocess.run(['conda', 'list', '--json'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return json.loads(result.stdout)
    except subprocess.CalledProcessError as e:
        return f"Error: {e.stderr}"


# Generate SciToken dynamically 
def generate_osdf_token(key_id, key_path, key_issuer) -> str:

    #for SciToken Generation & Validation
    import scitokens
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend

    dynamic_pass="" #Initialize Blank dynamic Password

    #Read Private Key using load_pem_private_key() method 
    if not os.path.exists(key_path):
        print(f"File {key_path} does not exist.")
        return dynamic_pass

    try:
        with open(key_path, "r") as file_pointer:
            private_key_contents = file_pointer.read()

        loaded_private_key = serialization.load_pem_private_key(
            private_key_contents.encode(),
            password=None, # Assumes Private key. Update this if password is used for Private Key
            backend=default_backend()
        )

        if is_url(key_issuer):
            token = scitokens.SciToken(key=loaded_private_key, key_id=key_id) #Generate SciToken
            #token.update_claims({"iss": key_issuer, "scope": "write:/ read:/", "aud": "NRP", "sub": "NRP"})
            token.update_claims({"scope": "write:/ read:/", "aud": "NRP", "sub": "NRP"}) #TODO: Figure out how to supply these as input params

            # Serialize the token to a string
            token_ser = token.serialize(issuer=key_issuer) 
            #Key_issuer is something like ""https://t.nationalresearchplatform.org/fdp"

            #Stringify token_str
            token_str=token_ser.decode()
            dynamic_pass="Bearer "+ token_str
        else:
            print(f"{key_issuer} is not a valid URL.")

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")

    return dynamic_pass


###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os

from cmflib import cmfquery
from cmflib.storage_backends import (
    minio_artifacts,
    local_artifacts,
    amazonS3_artifacts,
    sshremote_artifacts,
    osdf_artifacts,
)
from cmflib.cli.command import CmdBase
from cmflib.utils.dvc_config import DvcConfig


class CmdArtifactPull(CmdBase):

    def split_url_pipeline(self, url: str, pipeline_name: str):
       # This function takes url and pipeline_name as a input parameter
       # return string which contains the artifact repo path of the artifact
       # url = Test-env:/home/user/local-storage/files/md5/23/6d9502e0283d91f689d7038b8508a2
       # pipeline_name = Test-env

       # checking whether pipeline name exist inside url
       if pipeline_name in url:
            # if multiple pipelines logs same artifact, then spliting them using ',' delimiter 
            if "," in url:
                urls = url.split(",")
                # iterate over each urls
                for u in urls:
                    # assign u to url if pipeline name exist
                    if pipeline_name in u:
                        url = u
            # splitting url using ':' delimiter token = ["Test-env","home/user/local-storage/files/md5/23/6d9502e508a2"]
            token = url.split(":")
            # removing 1st element from token i.e pipeline name
            # output token will be ["home/user/local-storage/files/md5/23/6d9502e508a2"]
            token.pop(0)
            if len(token) > 1:
                # in case of metrics we have multiple ':' in its url
                # concating remaining tokens after removing pipeline_name using ':' delimiter
                token = ":".join(token)
                return token
            return "".join(token)

    def extract_repo_args(self, type: str, name: str, url: str, current_directory: str):
        # Extracting the repository URL, current path, bucket name, and other relevant
        # information from the user-supplied arguments.
        # url = Test-env:/home/user/local-storage/files/md5/06/d100ff3e04e2c87bf20f0feacc9034,
        #          Second-env:/home/user/local-storage/files/md5/06/d100ff3e04e2c"

        # s_url = Url without pipeline name
        s_url = self.split_url_pipeline(url, self.args.pipeline_name)

        # got url in the form of /home/user/local-storage/files/md5/06/d100ff3e04e2c
        #spliting url using '/' delimiter
        token = s_url.split("/")

        # name = artifacts/model/model.pkl
        name = name.split(":")[0]
        if type == "minio":
            token_length = len(token)

            # assigned 2nd position element to bucket_name
            bucket_name = token[2]

            # The folder structure of artifact data has been updated due to a change in the DVC 3.0 version
            # Previously, the structure was dvc-art/23/69v2uu3jeejjeiw
            # but now it includes additional directories and has become files dvc-art/files/md5/23/69v2uu3jeejjeiw.
            # Consequently, the previous logic takes only the last 2 elements from the list of tokens,
            # but with the new structure, it needs to take the last 4 elements.

            # get last 4 element inside token
            token = token[(token_length-4):]

            # join last 4 token using '/' delimiter
            object_name = "/".join(token)
            # output = files/md5/23/69v2uu3jeejjeiw

            path_name = current_directory + "/" + name
            return bucket_name, object_name, path_name

        elif type == "local":
            token_length = len(token)
            download_loc = current_directory + "/" + name

            # local artifact repo path =  local-storage/files/md5/23/69v2uu3jeejjeiw.

            # get last 4 element inside token
            token = token[(token_length-4):]

            # join last 4 token using '/' delimiter
            current_dvc_loc = "/".join(token)

            return current_dvc_loc, download_loc

        elif type == "ssh":
            # comments remaining
            token_var = token[2].split(":")
            host = token_var[0]
            token.pop(0)
            token.pop(0)
            token.pop(0)
            current_loc_1 = "/".join(token)
            current_loc = f"/{current_loc_1}"
            return host, current_loc, name
        elif type == "osdf":
            token_length = len(token)
            download_loc = current_directory + "/" + name if current_directory != ""  else name
            #current_dvc_loc = (token[(token_length - 2)] + "/" + token[(token_length - 1)])
            #return FQDNL of where to download from, where to download to, what the artifact will be named
            return s_url, download_loc, name
        else:
            # sometimes s_url is empty - this shouldn't happen technically
            # sometimes s_url is not starting with s3:// - technically this shouldn't happen
            if s_url and s_url.startswith("s3://"):
                url_with_bucket = s_url.split("s3://")[1]
                # url_with_bucket = mybucket/user/files/md5/23/6d9502e0283d91f689d7038b8508a2
                # splitting the string using '/' as the delimiter
                # bucket_name = mybucket
                # object_name = user/files/md5/23/6d9502e0283d91f689d7038b8508a2
                bucket_name, object_name = url_with_bucket.split('/', 1)
                download_loc =  current_directory + "/" + name if current_directory != ""  else name
                #print(download_loc)
                return bucket_name, object_name, download_loc
            else:
                # returning bucket_name, object_name and download_loc returning as empty
                return "", "", ""

    def search_artifact(self, input_dict):
        # This function takes input_dict as input artifact
        for name, url in input_dict.items():
            if not isinstance(url, str):
                continue
            # splitting name with ':' as the delimiter and store first argument inside name
            name = name.split(":")[0]
            file_name = name.split('/')[-1]
            if file_name == self.args.artifact_name:
                return name, url
            else:
                pass

    def run(self):
        # check whether the mlmd file exist or not in current directory
        pipeline_name = self.args.pipeline_name
        current_directory = os.getcwd()
        mlmd_file_name = "./mlmd"
        if self.args.file_name:
            mlmd_file_name = self.args.file_name
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
            current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):
            return f"ERROR: {mlmd_file_name} doesn't exists in {current_directory} directory."
        query = cmfquery.CmfQuery(mlmd_file_name)

        # getting all pipeline stages[i.e Prepare, Featurize, Train and Evaluate]
        stages = query.get_pipeline_stages(self.args.pipeline_name)
        executions = []
        identifiers = []

        for stage in stages:
            # getting all executions for stages
            executions = query.get_all_executions_in_stage(stage)
            # check if stage has executions
            if len(executions) > 0:
                 # converting it to dictionary
                dict_executions = executions.to_dict("dict")
                # append id's of executions inside identifiers
                for id in dict_executions["id"].values():
                    identifiers.append(id)
            else:
                print("No Executions found for " + stage + " stage.")

        # created dictionary
        name_url_dict = {}
        if len(identifiers) == 0:  # check if there are no executions
            return "No executions found."
        for identifier in identifiers:
            get_artifacts = query.get_all_artifacts_for_execution(
                identifier
            )  # getting all artifacts with id
            temp_dict = dict(zip(get_artifacts['name'], get_artifacts['url'])) # getting dictionary of name and url pair
            name_url_dict.update(temp_dict) # updating name_url_dict with temp_dict
        # print(name_url_dict)
        # name_url_dict = ('artifacts/parsed/test.tsv:6f597d341ceb7d8fbbe88859a892ef81', 'Test-env:/home/sharvark/local-storage/6f/597d341ceb7d8fbbe88859a892ef81'
        # name_url_dict = ('artifacts/parsed/test.tsv:6f597d341ceb7d8fbbe88859a892ef81', 'Test-env:/home/sharvark/local-storage/6f/597d341ceb7d8fbbe88859a892ef81,Second-env:/home/sharvark/local-storage/6f/597d341ceb7d8fbbe88859a892ef81')

        output = DvcConfig.get_dvc_config()  # pulling dvc config
        if type(output) is not dict:
            return output
        dvc_config_op = output

        if dvc_config_op["core.remote"] == "minio":
            minio_class_obj = minio_artifacts.MinioArtifacts()
            if self.args.artifact_name:
                output = self.search_artifact(name_url_dict)
                # output[0] = name
                # output[1] = url
                if output is None:
                    print(f"{self.args.artifact_name} doesn't exist.")
                else:
                    minio_args = self.extract_repo_args("minio", output[0], output[1], current_directory)
                    stmt = minio_class_obj.download_artifacts(
                        dvc_config_op,
                        current_directory,
                        minio_args[0], # bucket_name
                        minio_args[1], # object_name
                        minio_args[2], # path_name
                    )
                    print(stmt)
            else:
                for name, url in name_url_dict.items():
                    if not isinstance(url, str):
                        continue
                    minio_args = self.extract_repo_args("minio", name, url, current_directory)
                    stmt = minio_class_obj.download_artifacts(
                        dvc_config_op,
                        current_directory,
                        minio_args[0], # bucket_name
                        minio_args[1], # object_name
                        minio_args[2], # path_name
                    )
                    print(stmt)
            return "Done"
        elif dvc_config_op["core.remote"] == "local-storage":
            local_class_obj = local_artifacts.LocalArtifacts()
            if self.args.artifact_name:
                output = self.search_artifact(name_url_dict)
                # output[0] = name
                # output[1] = url
                if output is None:
                    print(f"{self.args.artifact_name} doesn't exist.")
                else:
                    local_args = self.extract_repo_args("local", output[0], output[1], current_directory)
                    stmt = local_class_obj.download_artifacts(
                           dvc_config_op, current_directory, local_args[0], local_args[1]
                    )
                    print(stmt)
            else:
                for name, url in name_url_dict.items():
                    #print(name, url)
                    if not isinstance(url, str):
                        continue
                    local_args = self.extract_repo_args("local", name, url, current_directory)
                    # local_args[0] = current dvc location
                    # local_args[1] = current download location
                    stmt = local_class_obj.download_artifacts(
                           dvc_config_op, current_directory, local_args[0], local_args[1]
                    )
                    print(stmt)
            return "Done"
        elif dvc_config_op["core.remote"] == "ssh-storage":
            sshremote_class_obj = sshremote_artifacts.SSHremoteArtifacts()
            if self.args.artifact_name:
                output = self.search_artifact(name_url_dict)
                # output[0] = name
                # output[1] = url
                if output is None:
                    print(f"{self.args.artifact_name} doesn't exist.")
                else:
                    args = self.extract_repo_args("ssh", output[0], output[1], current_directory)
                    stmt = sshremote_class_obj.download_artifacts(
                        dvc_config_op,
                        args[0], # host,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2]  # name
                    )
                    print(stmt)
            else:
                for name, url in name_url_dict.items():
                    #print(name, url)
                    if not isinstance(url, str):
                        continue
                    args = self.extract_repo_args("ssh", name, url, current_directory)
                    stmt = sshremote_class_obj.download_artifacts(
                        dvc_config_op,
                        args[0], # host,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2]  # name
                    )
                    print(stmt)
            return "Done"
        elif dvc_config_op["core.remote"] == "osdf":
            #Regenerate Token for OSDF
            from cmflib.utils.helper_functions import generate_osdf_token
            from cmflib.utils.helper_functions import is_url
            from cmflib.dvc_wrapper import dvc_add_attribute
            from cmflib.utils.cmf_config import CmfConfig
            #Fetch Config from CMF_Config_File
            cmf_config_file = os.environ.get("CONFIG_FILE", ".cmfconfig")
            cmf_config={}
            cmf_config=CmfConfig.read_config(cmf_config_file)
            #Regenerate password 
            dynamic_password = generate_osdf_token(cmf_config["osdf-key_id"],cmf_config["osdf-key_path"],cmf_config["osdf-key_issuer"])
            #cmf_config["password"]=dynamic_password
            #Update Password in .dvc/config for future use
            dvc_add_attribute(dvc_config_op["core.remote"],"password",dynamic_password)
            #Updating dvc_config_op data structure with new password as well since this is used in download_artifacts() below
            dvc_config_op["remote.osdf.password"]=dynamic_password
            #Need to write to cmfconfig with new credentials
            #CmfConfig.write_config(cmf_config, "osdf", attr_dict, True)
            #Now Ready to do dvc pull 

            osdfremote_class_obj = osdf_artifacts.OSDFremoteArtifacts()
            if self.args.artifact_name:
                output = self.search_artifact(name_url_dict)
                # output[0] = name
                # output[1] = url
                if output is None:
                    print(f"{self.args.artifact_name} doesn't exist.")
                else:
                    args = self.extract_repo_args("osdf", output[0], output[1], current_directory)
                    stmt = osdfremote_class_obj.download_artifacts(
                        dvc_config_op,
                        args[0], # s_url of the artifact
                        current_directory,
                        args[1], # download_loc of the artifact
                        args[2]  # name of the artifact
                    )
                    print(stmt)
            else:
                for name, url in name_url_dict.items():
                    #print(name, url)
                    if not isinstance(url, str):
                        continue
                    args = self.extract_repo_args("osdf", name, url, current_directory)
                    stmt = osdfremote_class_obj.download_artifacts(
                        dvc_config_op,
                        args[0], # host,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2]  # name
                    )
                    print(stmt)
            return "Done"
        elif dvc_config_op["core.remote"] == "amazons3":
            amazonS3_class_obj = amazonS3_artifacts.AmazonS3Artifacts()
            #print(self.args.artifact_name,"artifact name")
            if self.args.artifact_name:
                output = self.search_artifact(name_url_dict)
                # output[0] = name
                # output[1] = url
                if output is None:
                    print(f"{self.args.artifact_name} doesn't exist.")
                else:
                    args = self.extract_repo_args("amazons3", output[0], output[1], current_directory)
                    if args[0] and args[1] and args[2]:
                        stmt = amazonS3_class_obj.download_artifacts(
                            dvc_config_op,
                            current_directory,
                            args[0], # bucket_name
                            args[1], # object_name
                            args[2], # download_loc
                        )
                        print(stmt)
            else:
                for name, url in name_url_dict.items():
                    if not isinstance(url, str):
                        continue
                    args = self.extract_repo_args("amazons3", name, url, current_directory)
                    if args[0] and args[1] and args[2]:
                        stmt = amazonS3_class_obj.download_artifacts(
                            dvc_config_op,
                            current_directory,
                            args[0], # bucket_name
                            args[1], # object_name
                            args[2], # download_loc
                        )
                        print(stmt)
            return "Done"
        else:
            remote = dvc_config_op["core.remote"]
            msg = f"{remote} is not valid artifact repository for CMF.\n Reinitialize CMF."
            return msg


def add_parser(subparsers, parent_parser):
    PULL_HELP = "Pull artifacts from user configured repository."

    parser = subparsers.add_parser(
        "pull",
        parents=[parent_parser],
        description="Pull artifacts from user configured repository.",
        help=PULL_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", "--file_name", help="Specify mlmd file name.", metavar="<file_name>"
    )

    parser.add_argument(
        "-a", "--artifact_name", help="Specify artifact name.", metavar="<artifact_name>"
    )

    parser.set_defaults(func=CmdArtifactPull)




from setuptools import setup, find_packages

VERSION = '0.0.10'
DESCRIPTION = 'Metadata Python Package'
LONG_DESCRIPTION = 'Metadata framework storing AI metadata into MLMD'

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="cmflib",
        version=VERSION,
        author="Hewlett Packard Enterprise",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=["ml-metadata==1.15.0",
                          "dvc[ssh,s3]==3.51.1", "pandas", "retrying", "pyarrow", "neo4j", \
                            "scikit-learn", "tabulate", "click", "minio", "paramiko", "scitokens", "cryptography"],  # add any additional packages that
        # needs to be installed along with your package. Eg: 'caer'

        keywords=['python', 'first package'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 3",
            "Operating System :: POSIX :: Linux",
        ],
        scripts=['cmflib/bin/cmf']
)
